import uuid
import inspect
import ast
from typing import Any, AsyncGenerator, Callable, TYPE_CHECKING, overload

if TYPE_CHECKING:
    from elysia.tree.tree import Tree


class ToolMeta(type):
    """Metaclass that extracts tool metadata from __init__ method."""

    _tool_name: str | None = None
    _tool_description: str | None = None
    _tool_inputs: dict | None = None
    _tool_end: bool | None = None

    @staticmethod
    def _convert_ast_dict(ast_dict: ast.Dict) -> dict:
        out = {}
        for i in range(len(ast_dict.keys)):
            k: ast.Constant = ast_dict.keys[i]  # type: ignore
            v: ast.Dict | ast.List | ast.Constant = ast_dict.values[i]  # type: ignore

            if isinstance(v, ast.Dict):
                out[k.value] = ToolMeta._convert_ast_dict(v)
            elif isinstance(v, ast.List):
                out[k.value] = ToolMeta._convert_ast_list(v)
            elif isinstance(v, ast.Constant):
                out[k.value] = v.value
        return out

    @staticmethod
    def _convert_ast_list(ast_list: ast.List) -> list:
        out = []
        for v in ast_list.values:  # type: ignore
            if isinstance(v, ast.Dict):
                out.append(ToolMeta._convert_ast_dict(v))
            elif isinstance(v, ast.List):
                out.append(ToolMeta._convert_ast_list(v))
            elif isinstance(v, ast.Constant):
                out.append(v.value)
        return out

    def __new__(cls, name, bases, namespace):
        new_class = super().__new__(cls, name, bases, namespace)

        # Skip the base Tool class itself
        if name == "Tool":
            return new_class

        # Extract metadata from __init__ method if it exists
        if "__init__" in namespace:
            try:
                init_method = namespace["__init__"]
                source = inspect.getsource(init_method)
                tree = ast.parse(source.strip())

                # Find the super().__init__() call and extract arguments
                for node in ast.walk(tree):
                    if (
                        isinstance(node, ast.Call)
                        and isinstance(node.func, ast.Attribute)
                        and node.func.attr == "__init__"
                        and isinstance(node.func.value, ast.Call)
                        and isinstance(node.func.value.func, ast.Name)
                        and node.func.value.func.id == "super"
                    ):

                        # Extract keyword arguments
                        for keyword in node.keywords:
                            if keyword.arg == "name" and isinstance(
                                keyword.value, ast.Constant
                            ):
                                new_class._tool_name = keyword.value.value
                            elif keyword.arg == "description" and isinstance(
                                keyword.value, ast.Constant
                            ):
                                new_class._tool_description = keyword.value.value
                            elif keyword.arg == "inputs" and isinstance(
                                keyword.value, ast.Dict
                            ):
                                new_class._tool_inputs = ToolMeta._convert_ast_dict(
                                    keyword.value
                                )
                            elif keyword.arg == "end" and isinstance(
                                keyword.value, ast.Constant
                            ):
                                new_class._tool_end = keyword.value.value
                        break
            except Exception:
                # If parsing fails, just continue without setting class attributes
                pass

        return new_class


class Tool(metaclass=ToolMeta):
    """
    The generic Tool class, which will be a superclass of any tools used by Elysia.
    """

    @classmethod
    def get_metadata(cls) -> dict[str, str | bool | dict | None]:
        """Get tool metadata without instantiation."""
        return {
            "name": getattr(cls, "_tool_name", None),
            "description": getattr(cls, "_tool_description", None),
            "inputs": getattr(cls, "_tool_inputs", None),
            "end": getattr(cls, "_tool_end", False),
        }

    def __init__(
        self,
        name: str,
        description: str,
        status: str = "",
        inputs: dict = {},
        end: bool = False,
        **kwargs,
    ):
        """
        Args:
            name (str): The name of the tool. Required.
            description (str): A detailed description of the tool to give to the LLM. Required.
            status (str): A status update message to display while the tool is running. Optional, defaults to "Running {name}...".
            inputs (dict): A dictionary of inputs for the tool, with the following structure:
                ```python
                {
                    input_name: {
                        "description": str,
                        "type": str,
                        "default": str,
                        "required": bool
                    }
                }
                ```
            end (bool): Whether the tool is an end tool. Optional, defaults to False.
        """
        self.name = name
        self.description = description
        self.inputs = inputs
        self.end = end

        if status == "":
            self.status = f"Running {self.name}..."
        else:
            self.status = status

    def get_default_inputs(self) -> dict:
        return {
            key: value["default"] if "default" in value else None
            for key, value in self.inputs.items()
        }

    async def run_if_true(
        self,
        tree_data,
        base_lm,
        complex_lm,
        client_manager,
    ) -> tuple[bool, dict]:
        """
        This method is called to check if the tool should be run automatically.
        If this returns `True`, then the tool is immediately called at the start of the tree.
        Otherwise it does not appear in the decision tree.
        You must also return a dictionary of inputs for the tool, which will be used to call the tool if `True` is returned.
        If the inputs are None or an empty dictionary, then the default inputs will be used.

        This must be an async function.

        Args:
            tree_data (TreeData): The tree data object.
            base_lm (LM): The base language model, a dspy.LM object.
            complex_lm (LM): The complex language model, a dspy.LM object.
            client_manager (ClientManager): The client manager, a way of interfacing with a Weaviate client.

        Returns:
            bool: True if the tool should be run automatically, False otherwise.
            dict: A dictionary of inputs for the tool, with the following structure:
                ```python
                {
                    input_name: input_value
                }
                ```
        """
        return False, {}

    async def is_tool_available(
        self,
        tree_data,
        base_lm,
        complex_lm,
        client_manager,
    ) -> bool:
        """
        This method is called to check if the tool is available.
        If this returns `True`, then the tool is available to be used by the LLM.
        Otherwise it does not appear in the decision tree.

        The difference between this and `run_if_true` is that `run_if_true` will _always_ run the __call__ method if it returns `True`,
        even if the LLM does not choose to use the tool.

        This must be an async function.

        Args:
            tree_data (TreeData): The tree data object.
            base_lm (LM): The base language model, a dspy.LM object.
            complex_lm (LM): The complex language model, a dspy.LM object.
            client_manager (ClientManager): The client manager, a way of interfacing with a Weaviate client.

        Returns:
            bool: True if the tool is available, False otherwise.
        """
        return True

    async def __call__(
        self, tree_data, inputs, base_lm, complex_lm, client_manager, **kwargs
    ) -> AsyncGenerator[Any, None]:
        """
        This method is called to run the tool.

        This must be an async generator, so objects must be yielded and not returned.

        Args:
            tree_data (TreeData): The data from the decision tree, includes the environment, user prompt, etc.
                See the `TreeData` class for more details.
            base_lm (LM): The base language model, a dspy.LM object.
            complex_lm (LM): The complex language model, a dspy.LM object.
            client_manager (ClientManager): The client manager, a way of interfacing with a Weaviate client.
        """
        yield None


@overload
def tool(
    function: Callable,
    *,
    status: str | None = None,
    end: bool = False,
    tree: "Tree | None" = None,
    branch_id: str | None = None,
) -> Tool: ...


@overload
def tool(
    function: None = None,
    *,
    status: str | None = None,
    end: bool = False,
    tree: "Tree | None" = None,
    branch_id: str | None = None,
) -> Callable[[Callable], Tool]: ...


def tool(
    function: Callable | None = None,
    *,
    status: str | None = None,
    end: bool = False,
    tree: "Tree | None" = None,
    branch_id: str | None = None,
) -> Callable[[Callable], Tool] | Tool:
    """
    Create a tool from a python function.
    Use this decorator to create a tool from a function.
    The function must be an async function or async generator function.

    Args:
        function (Callable): The function to create a tool from.
        status (str | None): The status message to display while the tool is running.
            Optional, defaults to None, which will use the default status message "Running {tool_name}...".
        end (bool): Whether the tool can be at the end of the decision tree.
            Set to True when this tool is allowed to end the conversation.
            Optional, defaults to False.
        tree (Tree | None): The tree to add the tool to.
            Optional, defaults to None, which will not add the tool to the tree.
        branch_id (str | None): The branch of the tree to add the tool to.
            Optional, defaults to None, which will add the tool to the root branch.

    Returns:
        (Tool): The tool object which can be added to the tree (via `tree.add_tool(...)`).
    """

    def decorator(function: Callable) -> Tool:
        async_function = inspect.iscoroutinefunction(function)
        async_generator_function = inspect.isasyncgenfunction(function)

        if not async_function and not async_generator_function:
            raise TypeError(
                "The provided function must be an async function or async generator function."
            )

        if "inputs" in list(function.__annotations__.keys()):
            raise TypeError(
                "The `inputs` argument is reserved in tool functions, please choose another name."
            )

        sig = inspect.signature(function)
        defaults_mapping = {
            k: v.default
            for k, v in sig.parameters.items()
            if v.default is not inspect.Parameter.empty
        }

        def list_to_list_of_dicts(result: list) -> list[dict]:
            objects = []
            for obj in result:
                if isinstance(obj, dict):
                    objects.append(obj)
                elif isinstance(obj, int | float | bool):
                    objects.append(
                        {
                            "tool_result": obj,
                        }
                    )
                elif isinstance(obj, list):
                    objects.append(list_to_list_of_dicts(obj))
                else:
                    objects.append(
                        {
                            "tool_result": obj,
                        }
                    )
            return objects

        def return_mapping(result, inputs: dict):
            if isinstance(result, Result | Text | Update | Status | Error):
                return result
            elif isinstance(result, str):
                return Response(result)
            elif isinstance(result, int | float | bool):
                return Result(
                    objects=[
                        {
                            "tool_result": result,
                        }
                    ],
                    metadata={
                        "tool_name": function.__name__,
                        "inputs_used": inputs,
                    },
                )
            elif isinstance(result, dict):
                return Result(
                    objects=[result],
                    metadata={
                        "tool_name": function.__name__,
                        "inputs_used": inputs,
                    },
                )
            elif isinstance(result, list):
                return Result(
                    objects=list_to_list_of_dicts(result),
                    metadata={
                        "tool_name": function.__name__,
                        "inputs_used": inputs,
                    },
                )
            else:
                return Result(
                    objects=[
                        {
                            "tool_result": result,
                        }
                    ],
                    metadata={
                        "tool_name": function.__name__,
                        "inputs_used": inputs,
                    },
                )

        class ToolClass(Tool):
            def __init__(self, **kwargs):
                super().__init__(
                    name=function.__name__,
                    description=function.__doc__ or "",
                    status=(
                        status
                        if status is not None
                        else f"Running {function.__name__}..."
                    ),
                    inputs={
                        input_key: {
                            "description": "",
                            "type": input_value,
                            "default": defaults_mapping.get(input_key, None),
                            "required": defaults_mapping.get(input_key, None)
                            is not None,
                        }
                        for input_key, input_value in function.__annotations__.items()
                        if input_key
                        not in [
                            "tree_data",
                            "base_lm",
                            "complex_lm",
                            "client_manager",
                            "return",
                        ]
                    },
                    end=end,
                )
                self._original_function = function

            async def __call__(
                self, tree_data, inputs, base_lm, complex_lm, client_manager, **kwargs
            ):

                if async_function:
                    tool_output = [
                        await function(
                            **inputs,
                            **{
                                k: v
                                for k, v in {
                                    "tree_data": tree_data,
                                    "base_lm": base_lm,
                                    "complex_lm": complex_lm,
                                    "client_manager": client_manager,
                                    **kwargs,
                                }.items()
                                if k in function.__annotations__
                            },
                        )
                    ]
                elif async_generator_function:
                    results = []
                    async for result in function(
                        **inputs,
                        **{
                            k: v
                            for k, v in {
                                "tree_data": tree_data,
                                "base_lm": base_lm,
                                "complex_lm": complex_lm,
                                "client_manager": client_manager,
                                **kwargs,
                            }.items()
                            if k in function.__annotations__
                        },
                    ):
                        results.append(result)
                    tool_output = results

                for result in tool_output:
                    mapped_result = return_mapping(result, inputs)
                    yield mapped_result

        tool_class = ToolClass()

        if tree is not None:
            tree.add_tool(tool_class, branch_id=branch_id)

        return tool_class

    if function is not None:
        return decorator(function)
    else:
        return decorator


class Return:
    """
    A Return object is something that is returned to the frontend.
    This kind of object is frontend-aware.
    """

    def __init__(self, frontend_type: str, payload_type: str):
        self.frontend_type = frontend_type  # frontend identifier
        self.payload_type = payload_type  # backend identifier


class Text(Return):
    """
    Frontend Return Type 1: Text
    Objects is usually a one-element list containing a dict with a "text" key only.
    But is not limited to this.
    """

    def __init__(
        self,
        payload_type: str,
        objects: list[dict],
        metadata: dict = {},
        display: bool = True,
    ):
        Return.__init__(self, "text", payload_type)
        self.objects = objects
        self.metadata = metadata
        self.text = self._concat_text(self.objects)
        self.display = display

    def _concat_text(self, objects: list[dict]):
        text = ""
        for i, obj in enumerate(objects):
            if "text" in obj:
                spacer = ""
                if (
                    not obj["text"].endswith(" ")
                    and not obj["text"].endswith("\n")
                    and i != len(objects) - 1
                ):
                    spacer += " "

                if (
                    i != len(objects) - 1
                    and "text" in objects[i + 1]
                    and objects[i + 1]["text"].startswith("* ")
                    and not obj["text"].endswith("\n")
                ):
                    spacer += "\n"

                text += obj["text"] + spacer

        text = text.replace("_REF_ID", "")
        text = text.replace("REF_ID", "")
        text = text.replace("  ", " ")
        return text

    def to_json(self):
        return {
            "type": self.payload_type,
            "metadata": self.metadata,
            "objects": self.objects,
        }

    async def to_frontend(
        self,
        user_id: str,
        conversation_id: str,
        query_id: str,
    ):
        if not self.display:
            return

        return {
            "type": self.frontend_type,
            "id": self.frontend_type[:3] + "-" + str(uuid.uuid4()),
            "user_id": user_id,
            "conversation_id": conversation_id,
            "query_id": query_id,
            "payload": self.to_json(),
        }


class Response(Text):
    def __init__(self, text: str, **kwargs):
        Text.__init__(self, "response", [{"text": text}], **kwargs)


class Update(Return):
    """
    Frontend Return Type 2: Update
    An update is something that is not _displayed_ on the frontend, but gives information to the frontend.
    E.g. a warning, error, status message, etc.
    """

    def __init__(self, frontend_type: str, object: dict):
        Return.__init__(self, frontend_type, "update")
        self.object = object

    def to_json(self):
        return self.object

    async def to_frontend(self, user_id: str, conversation_id: str, query_id: str):
        return {
            "type": self.frontend_type,
            "user_id": user_id,
            "conversation_id": conversation_id,
            "query_id": query_id,
            "id": self.frontend_type[:3] + "-" + str(uuid.uuid4()),
            "payload": self.to_json(),
        }


class Status(Update):
    """
    Status message to be sent to the frontend for real-time updates in words.
    """

    def __init__(self, text: str):
        self.text = text
        Update.__init__(self, "status", {"text": text})


class Warning(Update):
    """
    Warning message to be sent to the frontend.
    """

    def __init__(self, text: str):
        self.text = text
        Update.__init__(self, "warning", {"text": text})


class Completed(Update):
    """
    Completed message to be sent to the frontend (tree is complete all recursions).
    """

    def __init__(self):
        Update.__init__(self, "completed", {})


class Result(Return):
    """
    Result objects are returned to the frontend.
    These are displayed on the frontend.
    E.g. a table, a chart, a text response, etc.

    You can yield a `Result` directly, and specify the `type` and `name` of the result.
    """

    def __init__(
        self,
        objects: list[dict],
        metadata: dict = {},
        payload_type: str = "default",
        name: str = "default",
        mapping: dict | None = None,
        llm_message: str | None = None,
        unmapped_keys: list[str] = ["_REF_ID"],
        display: bool = True,
    ):
        """
        Args:
            objects (list[dict]): The objects attached to this result.
            payload_type: (str): Identifier for the type of result.
            metadata (dict): The metadata attached to this result,
                for example, query used, time taken, any other information not directly within the objects
            name (str): The name of the result, e.g. this could be the name of the collection queried.
                Used to index the result in the environment.
            mapping (dict | None): A mapping of the objects to the frontend.
                Essentially, if the objects are going to be displayed on the frontend, the frontend has specific keys that it wants the objects to have.
                This is a dictionary that maps from frontend keys to the object keys.
            llm_message (str | None): A message to be displayed to the LLM to help it parse the result.
                You can use the following placeholders that will automatically be replaced with the correct values:

                - {type}: The type of the object
                - {name}: The name of the object
                - {num_objects}: The number of objects in the object
                - {metadata_key}: Any key in the metadata dictionary
            unmapped_keys (list[str]): A list of keys that are not mapped to the frontend.
            display (bool): Whether to display the result on the frontend when yielding this object.
                Defaults to `True`.
        """
        Return.__init__(self, "result", payload_type)
        self.objects = objects
        self.metadata = metadata
        self.name = name
        self.mapping = mapping
        self.llm_message = llm_message
        self.unmapped_keys = unmapped_keys
        self.display = display

    def __len__(self):
        return len(self.objects)

    def format_llm_message(self):
        """
        llm_message is a string that is used to help the LLM parse the output of the tool.
        It is a placeholder for the actual message that will be displayed to the user.

        You can use the following placeholders:

        - {payload_type}: The type of the object
        - {name}: The name of the object
        - {num_objects}: The number of objects in the object
        - {metadata_key}: Any key in the metadata dictionary
        """
        if self.llm_message is None:
            return ""

        return self.llm_message.format_map(
            {
                "payload_type": self.payload_type,
                "name": self.name,
                "num_objects": len(self),
                **self.metadata,
            }
        )

    def do_mapping(self, objects: list[dict]):

        if self.mapping is None:
            return objects

        output_objects = []
        for obj in objects:
            output_objects.append(
                {
                    key: obj[self.mapping[key]]
                    for key in self.mapping
                    if self.mapping[key] != ""
                }
            )
            output_objects[-1].update(
                {key: obj[key] for key in self.unmapped_keys if key in obj}
            )

        return output_objects

    def to_json(self, mapping: bool = False):
        """
        Convert the result to a list of dictionaries.

        Args:
            mapping (bool): Whether to map the objects to the frontend.
                This will use the `mapping` dictionary created on initialisation of the `Result` object.
                If `False`, the objects are returned as is.
                Defaults to `False`.

        Returns:
            (list[dict]): A list of dictionaries, which can be serialised to JSON.
        """
        from elysia.util.parsing import format_dict_to_serialisable

        assert all(
            isinstance(obj, dict) for obj in self.objects
        ), "All objects must be dictionaries"

        if mapping and self.mapping is not None:
            output_objects = self.do_mapping(self.objects)
        else:
            output_objects = self.objects

        for object in output_objects:
            format_dict_to_serialisable(object)

        return output_objects

    async def to_frontend(
        self,
        user_id: str,
        conversation_id: str,
        query_id: str,
    ):
        """
        Convert the result to a frontend payload.
        This is a wrapper around the `to_json` method.
        But the objects and metadata are inside a `payload` key, which also includes a `type` key,
        being the frontend identifier for the type of payload being sent.
        (e.g. `ticket`, `product`, `message`, etc.)

        The outside of the payload also contains the user_id, conversation_id, and query_id.

        Args:
            user_id (str): The user ID.
            conversation_id (str): The conversation ID.
            query_id (str): The query ID.

        Returns:
            (dict): The frontend payload, which is a dictionary with the following structure:
                ```python
                {
                    "type": "result",
                    "user_id": str,
                    "conversation_id": str,
                    "query_id": str,
                    "id": str,
                    "payload": dict = {
                        "type": str,
                        "objects": list[dict],
                        "metadata": dict,
                    },
                }
                ```
        """
        if not self.display:
            return

        objects = self.to_json(mapping=True)
        if len(objects) == 0:
            return

        payload = {
            "type": self.payload_type,
            "objects": objects,
            "metadata": self.metadata,
        }

        return {
            "type": self.frontend_type,
            "user_id": user_id,
            "conversation_id": conversation_id,
            "query_id": query_id,
            "id": self.frontend_type[:3] + "-" + str(uuid.uuid4()),
            "payload": payload,
        }

    def llm_parse(self):
        """
        This method is called when the result is displayed to the LLM.
        It is used to display custom information to the LLM about the result.
        If `llm_message` was defined, then the llm message is formatted using the placeholders.
        Otherwise a default message is used.

        Returns:
            (str): The formatted llm message.
        """

        if self.llm_message is not None:
            return self.format_llm_message()
        else:
            # Default message
            return f"Displayed: A {self.payload_type} object with {len(self.objects)} objects."


# Retrieval/Aggregation/Anything we used code to get data
class Retrieval(Result):
    """
    Store of returned objects from a query/aggregate/any displayed results.
    """

    def __init__(
        self,
        objects: list[dict],
        metadata: dict = {},
        payload_type: str = "default",
        name: str | None = None,
        mapping: dict | None = None,
        unmapped_keys: list[str] = [
            "uuid",
            "ELYSIA_SUMMARY",
            "collection_name",
            "_REF_ID",
        ],
        display: bool = True,
    ) -> None:
        if name is None and "collection_name" in metadata:
            result_name = metadata["collection_name"]
        elif name is None:
            result_name = "default"
        else:
            result_name = name

        Result.__init__(
            self,
            objects=objects,
            payload_type=payload_type,
            metadata=metadata,
            name=result_name,
            mapping=mapping,
            unmapped_keys=unmapped_keys,
            display=display,
        )

    def add_summaries(self, summaries: list[str] = []) -> None:
        for i, obj in enumerate(self.objects):
            if i < len(summaries):
                obj["ELYSIA_SUMMARY"] = summaries[i]
            else:
                obj["ELYSIA_SUMMARY"] = ""

    def llm_parse(self) -> str:
        out = ""
        count = len(self.objects)

        if "collection_name" in self.metadata:
            if count != 0:
                out += f"\nQueried collection: '{self.metadata['collection_name']}' and returned {count} objects, "
            else:
                out += f"\nQueried collection: '{self.metadata['collection_name']}' but no objects were returned."
                out += f" Since it had no objects, judge the query that was created, and evaluate whether it was appropriate for the collection, the user prompt, and the data available."
                out += f" If it seemed innappropriate, you can choose to try again if you think it can still be completed (or there is more to do)."
        if "return_type" in self.metadata:
            out += f"\nReturned with type '{self.metadata['return_type']}', "
        if "output_type" in self.metadata:
            out += f"\noutputted '{'itemised summaries' if self.metadata['output_type'] == 'summary' else 'original objects'}'\n"
        if "query_text" in self.metadata:
            out += f"\nSearch terms: '{self.metadata['query_text']}'"
        if "query_type" in self.metadata:
            out += f"\nType of query: '{self.metadata['query_type']}'"
        if "impossible" in self.metadata:
            out += f"\nImpossible prompt: '{self.metadata['impossible']}'"
            if "collection_name" in self.metadata:
                out += f"\nThis attempt at querying the collection: {self.metadata['collection_name']} was deemed impossible."
            if "impossible_reason" in self.metadata:
                out += f"\nReasoning for impossibility: {self.metadata['impossible_reason']}"
        if "query_output" in self.metadata:
            out += f"\nThe query used was:\n{self.metadata['query_output']}"
        return out

    async def to_frontend(
        self,
        user_id: str,
        conversation_id: str,
        query_id: str,
    ) -> dict | None:
        objects = self.to_json(mapping=True)
        if len(objects) == 0:
            return

        payload = {
            "type": self.payload_type,
            "objects": objects,
            "metadata": self.metadata,
        }

        if "code" in self.metadata:
            payload["code"] = self.metadata["code"]

        return {
            "type": self.frontend_type,
            "user_id": user_id,
            "conversation_id": conversation_id,
            "query_id": query_id,
            "id": self.frontend_type[:3] + "-" + str(uuid.uuid4()),
            "payload": payload,
        }


class Error(Update):
    """
    Error objects are used to communicate errors to the decision agent/tool calls.
    When yielded, Error objects are automatically saved inside the TreeData object.
    When calling the same tool again, the saved Error object is automatically loaded into any tool calls made with the same tool name.
    All errors are shown to the decision agent to help decide whether the tool should be called again (retried), or a different tool should be called.
    """

    def __init__(self, feedback: str = "", error_message: str = ""):
        """
        Args:
            feedback (str): The feedback to display to the decision agent.
                Usually this will be the error message, but it could be formatted more specifically.
        """
        self.feedback = feedback
        self.error_message = error_message

        if feedback == "":
            self.feedback = "An unknown issue occurred."

        Update.__init__(
            self,
            "self_healing_error",
            {"feedback": self.feedback, "error_message": self.error_message},
        )



import asyncio
import nest_asyncio
import inspect
import types


def _to_task(future, as_task, loop):
    if not as_task or isinstance(future, asyncio.Task):
        return future
    return loop.create_task(future)


def asyncio_run(future, as_task=True):
    """
    From: https://stackoverflow.com/questions/46827007/error-runtimeerror-this-event-loop-is-already-running-in-python

    A better implementation of `asyncio.run`.

    :param future: A future or task or call of an async method.
    :param as_task: Forces the future to be scheduled as task (needed for e.g. aiohttp).
    """

    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:  # No event loop running:
        loop = asyncio.new_event_loop()
        return loop.run_until_complete(_to_task(future, as_task, loop))
    else:
        nest_asyncio.apply(loop)
        return asyncio.run(_to_task(future, as_task, loop))  # type: ignore



import pytest
import json
from fastapi.responses import JSONResponse


from elysia.api.core.log import set_log_level
from elysia.api.dependencies.common import get_user_manager

set_log_level("CRITICAL")

from elysia.api.routes.init import initialise_tree
from elysia.api.routes.utils import (
    follow_up_suggestions,
)
from elysia.api.routes.init import initialise_user
from elysia.api.api_types import (
    FollowUpSuggestionsData,
    InitialiseTreeData,
)


def read_response(response: JSONResponse):
    return json.loads(response.body)


class fake_websocket:
    results = []

    async def send_json(self, data: dict):
        self.results.append(data)


class TestUtils:

    @pytest.mark.asyncio
    async def test_follow_up_suggestions(self):
        user_id = "test_user_follow_up_suggestions"
        conversation_id = "test_conversation_follow_up_suggestions"

        user_manager = get_user_manager()

        out = await initialise_user(
            user_id,
            user_manager,
        )
        response = read_response(out)
        assert response["error"] == ""

        out = await initialise_tree(
            user_id,
            conversation_id,
            InitialiseTreeData(
                low_memory=True,
            ),
            user_manager,
        )
        response = read_response(out)
        assert response["error"] == ""

        out = await follow_up_suggestions(
            FollowUpSuggestionsData(
                user_id=user_id,
                conversation_id=conversation_id,
            ),
            user_manager,
        )

        response = read_response(out)
        assert response["error"] == ""


if __name__ == "__main__":
    import asyncio

    asyncio.run(TestUtils().test_follow_up_suggestions())



import pytest
from elysia.config import Settings
from elysia.tree.tree import Tree
from elysia.util.client import ClientManager


def test_tree_init():
    tree = Tree()
    assert tree is not None


def test_tree_init_with_settings():
    settings = Settings.from_smart_setup()
    tree = Tree(settings=settings)
    assert tree is not None


def test_run_tree():
    settings = Settings()
    settings.configure(
        base_model="gpt-4o-mini",
        base_provider="openai",
        complex_model="gpt-4o",
        complex_provider="openai",
    )
    tree = Tree(settings=settings)
    response, objects = tree.run("Hi!")
    assert response is not None and len(response) > 0
    assert len(objects) is not None


@pytest.mark.asyncio
async def test_init_tree_with_no_client():
    settings = Settings()
    settings.configure(
        base_model="gpt-4o-mini",
        base_provider="openai",
        complex_model="gpt-4o",
        complex_provider="openai",
    )
    settings.WCD_URL = ""
    settings.WCD_API_KEY = ""

    tree = Tree(settings=settings)

    assert not (
        await tree.tools["query"].is_tool_available(
            tree_data=tree.tree_data,
            base_lm=tree.base_lm,
            complex_lm=tree.complex_lm,
            client_manager=ClientManager(
                wcd_url=settings.WCD_URL, wcd_api_key=settings.WCD_API_KEY
            ),
        )
    )
    assert not (
        await tree.tools["aggregate"].is_tool_available(
            tree_data=tree.tree_data,
            base_lm=tree.base_lm,
            complex_lm=tree.complex_lm,
            client_manager=ClientManager(
                wcd_url=settings.WCD_URL, wcd_api_key=settings.WCD_API_KEY
            ),
        )
    )



# self-healing errors can be tested by making a custom tool that returns an error only on thef irst
# and then returns a result on the second call
from elysia import Tree, Tool, Error
from elysia.tree.objects import TreeData
from elysia.util.client import ClientManager
import dspy


class TestSelfHealing(Tool):

    def __init__(self, **kwargs):
        super().__init__(
            name="test_self_healing",
            description="Test self-healing",
            status="healing...",
            inputs={},
            end=False,
            **kwargs,
        )
        self.first_call = True

    async def __call__(
        self,
        tree_data: TreeData,
        inputs: dict,
        base_lm: dspy.LM,
        complex_lm: dspy.LM,
        client_manager: ClientManager,
        **kwargs,
    ):
        if self.first_call:
            yield Error("The first call of this tool _always_ fails.")
            self.first_call = False


def test_self_healing():
    tree = Tree()
    tree.add_tool(TestSelfHealing)

    tree("Call the self-healing tool")
    assert "error" in tree.tree_data.tasks_completed[0]["task"][0]
    assert tree.tree_data.tasks_completed[0]["task"][0]["error"]
    assert "error" not in tree.tree_data.tasks_completed[0]["task"][1]



import pytest
import inspect

from elysia.tools.retrieval.chunk import AsyncCollectionChunker, Chunker
from elysia.util.client import ClientManager

from weaviate.classes.query import QueryReference
from weaviate.classes.config import Configure, DataType, Property, ReferenceProperty
from weaviate.collections.classes.data import DataObject, DataReference
from weaviate.collections import CollectionAsync
from weaviate.collections.classes.internal import Object, QueryReturn
from weaviate.collections.classes.config_vectorizers import _Vectorizer
from weaviate.exceptions import WeaviateInvalidInputError
from weaviate.util import generate_uuid5
from weaviate.client import WeaviateAsyncClient


def test_chunker():
    chunker = Chunker(chunking_strategy="sentences", num_sentences=1)
    doc = "Hello, world! This is a test."
    chunks, spans = chunker.chunk(doc)
    assert len(chunks) == 2
    assert spans == [(0, 13), (14, 29)]
    assert chunks[0] == "Hello, world!"
    assert chunks[1] == "This is a test."
    assert doc[spans[0][0] : spans[0][1]] == chunks[0]
    assert doc[spans[1][0] : spans[1][1]] == chunks[1]

    doc = "Hello, world! This is a test. This is another test."
    chunks, spans = chunker.chunk(doc)
    assert len(chunks) == 3
    assert spans == [(0, 13), (14, 29), (30, 51)]
    assert chunks[0] == "Hello, world!"
    assert chunks[1] == "This is a test."
    assert chunks[2] == "This is another test."
    assert doc[spans[0][0] : spans[0][1]] == chunks[0]
    assert doc[spans[1][0] : spans[1][1]] == chunks[1]
    assert doc[spans[2][0] : spans[2][1]] == chunks[2]



import asyncio

from fastapi import APIRouter, Depends
from fastapi.websockets import WebSocket, WebSocketDisconnect
from typing import Callable
import psutil
import time

# Logging
from elysia.api.core.log import logger

# User manager
from elysia.api.dependencies.common import get_user_manager
from elysia.api.services.user import UserManager

# Websocket
from elysia.api.utils.websocket import help_websocket

# Preprocessing
from elysia.preprocessing.collection import preprocess_async

router = APIRouter()


async def process_collection(
    data: dict, websocket: WebSocket, user_manager: UserManager
):
    logger.debug(f"/process_collection API request received")
    logger.debug(f"User ID: {data['user_id']}")
    logger.debug(f"Collection name: {data['collection_name']}")

    user = await user_manager.get_user_local(data["user_id"])
    settings = user["tree_manager"].settings

    async for result in preprocess_async(
        collection_name=data["collection_name"],
        client_manager=user["client_manager"],
        force=True,
        settings=settings,
    ):
        try:
            logger.debug(
                f"(process_collection) sending result with progress: {result['progress']*100}%"
            )
            await websocket.send_json(result)
        except WebSocketDisconnect:
            logger.info("Client disconnected during process_collection")
            break
        await asyncio.sleep(0.001)
    logger.debug(f"(process_collection) FINISHED!")


@router.websocket("/process_collection")
async def process_collection_websocket(
    websocket: WebSocket, user_manager: UserManager = Depends(get_user_manager)
):
    await help_websocket(
        websocket, lambda data, ws: process_collection(data, ws, user_manager)
    )



import pytest
import os
import dotenv

from elysia.api.services.user import UserManager
from elysia.api.services.tree import TreeManager
from elysia.config import Settings
from elysia.api.utils.config import Config

dotenv.load_dotenv(override=True)


@pytest.mark.asyncio
async def test_config_in_managers():
    """
    Test updating settings and API keys in the user manager and tree manager.
    """
    user_id = "test_user_config"
    conversation_id = "test_conversation_config"

    user_manager = UserManager()

    # add a tree manager to the user manager without a settings
    await user_manager.add_user_local(user_id)

    # add a tree to the user manager
    await user_manager.initialise_tree(user_id, conversation_id, low_memory=True)

    # get tree manager
    tree_manager = user_manager.users[user_id]["tree_manager"]

    # check that settings is equal to environment variables
    assert tree_manager.settings.API_KEYS["openai_api_key"] == os.environ.get(
        "OPENAI_API_KEY"
    )
    assert tree_manager.settings.WCD_URL == os.environ.get("WCD_URL")
    assert tree_manager.settings.WCD_API_KEY == os.environ.get("WCD_API_KEY")

    # check that the tree has these settings too
    tree = await user_manager.get_tree(user_id, conversation_id)
    assert tree.settings.API_KEYS["openai_api_key"] == os.environ.get("OPENAI_API_KEY")
    assert tree.settings.WCD_URL == os.environ.get("WCD_URL")
    assert tree.settings.WCD_API_KEY == os.environ.get("WCD_API_KEY")

    # create a new user with a new config
    user_id_2 = "test_user_config_2"
    conversation_id_2 = "test_conversation_config_2"

    new_settings = Settings()
    new_settings.configure(
        openai_api_key="new_openai_api_key",
        wcd_url=os.getenv("WCD_URL"),
        wcd_api_key=os.getenv("WCD_API_KEY"),
        base_model="gpt-4o-mini",
        base_provider="openai",
        complex_model="gpt-4o",
        complex_provider="openai",
    )

    await user_manager.add_user_local(
        user_id_2,
        config=Config(
            id="new_config",
            name="New Config",
            settings=new_settings,
            style="Informative, polite and friendly.",
            agent_description="You search and query Weaviate to satisfy the user's query, providing a concise summary of the results.",
            end_goal=(
                "You have satisfied the user's query, and provided a concise summary of the results. "
                "Or, you have exhausted all options available, or asked the user for clarification."
            ),
            branch_initialisation="one_branch",
        ),
    )
    await user_manager.initialise_tree(user_id_2, conversation_id_2, low_memory=True)

    # get tree manager
    tree_manager_2 = user_manager.users[user_id_2]["tree_manager"]

    # check that settings is equal to new settings
    assert tree_manager_2.settings.API_KEYS["openai_api_key"] == "new_openai_api_key"
    assert tree_manager_2.settings.WCD_URL == os.getenv("WCD_URL")
    assert tree_manager_2.settings.WCD_API_KEY == os.getenv("WCD_API_KEY")

    # check that the tree has these settings too
    tree_2 = await user_manager.get_tree(user_id_2, conversation_id_2)
    assert tree_2.settings.API_KEYS["openai_api_key"] == "new_openai_api_key"
    assert tree_2.settings.WCD_URL == os.getenv("WCD_URL")
    assert tree_2.settings.WCD_API_KEY == os.getenv("WCD_API_KEY")



from elysia.tree.tree import Tree
from elysia.tree.util import get_saved_trees_weaviate
from elysia.util.client import ClientManager
from elysia.config import Settings
from elysia.preprocessing import preprocess_async
from elysia.preprocessing.collection import view_preprocessed_collection
import pytest

from deepeval import evaluate, metrics
from deepeval.test_case import LLMTestCase, LLMTestCaseParams
from weaviate.util import generate_uuid5
import weaviate.classes as wvc


def create_regular_vectorizer_collection(
    client_manager: ClientManager, collection_name: str
):

    with client_manager.connect_to_client() as client:

        if client.collections.exists(collection_name):
            client.collections.delete(collection_name)

        collection = client.collections.create(
            name=collection_name,
            properties=[
                wvc.config.Property(
                    name="product_id", data_type=wvc.config.DataType.TEXT
                ),
                wvc.config.Property(name="title", data_type=wvc.config.DataType.TEXT),
                wvc.config.Property(
                    name="description", data_type=wvc.config.DataType.TEXT
                ),
                wvc.config.Property(
                    name="tags", data_type=wvc.config.DataType.TEXT_ARRAY
                ),
                wvc.config.Property(name="price", data_type=wvc.config.DataType.NUMBER),
            ],
            vectorizer_config=wvc.config.Configure.Vectorizer.text2vec_openai(
                vectorize_collection_name=False
            ),
        )

        products_data = [
            {
                "product_id": "prod1",
                "title": "Wireless Bluetooth Headphones",
                "description": "High-quality wireless headphones with noise cancellation and 30-hour battery life.",
                "tags": ["electronics", "audio", "wireless", "bluetooth"],
                "price": 199.99,
            },
            {
                "product_id": "prod2",
                "title": "Organic Green Tea",
                "description": "Premium organic green tea sourced from high-altitude tea gardens in Japan.",
                "tags": ["beverage", "tea", "organic", "healthy"],
                "price": 24.99,
            },
            {
                "product_id": "prod3",
                "title": "Yoga Mat with Alignment Lines",
                "description": "Non-slip yoga mat with helpful alignment lines for proper posture during practice.",
                "tags": ["fitness", "yoga", "exercise", "wellness"],
                "price": 39.99,
            },
            {
                "product_id": "prod4",
                "title": "Smart Home Security Camera",
                "description": "WiFi-enabled security camera with motion detection and smartphone notifications.",
                "tags": ["security", "smart-home", "camera", "surveillance"],
                "price": 149.99,
            },
        ]

        with collection.batch.fixed_size(batch_size=10) as batch:
            for product in products_data:
                batch.add_object(
                    properties={
                        "title": product["title"],
                        "description": product["description"],
                        "tags": product["tags"],
                        "price": product["price"],
                    },
                    uuid=generate_uuid5(product["product_id"]),
                )


@pytest.mark.asyncio
async def test_save_load_weaviate():

    client_manager = ClientManager()
    collection_name = "Test_ELYSIA_Product_Data"
    create_regular_vectorizer_collection(client_manager, collection_name)

    # preprocess
    async for result in preprocess_async(collection_name, client_manager, force=True):
        if "error" in result and result["error"] != "":
            assert False, result["error"]

    try:
        settings = Settings.from_smart_setup()
        tree = Tree(settings=settings)

        # ask about edward
        response, objects = tree(
            "What is the description of the Yoga Mat with Alignment Lines",
            collection_names=[
                collection_name,
            ],
        )

        # save the tree to weaviate
        await tree.export_to_weaviate(
            collection_name="Test_ELYSIA_save_load_collection_llm",
            client_manager=client_manager,
        )

        # load the tree from weaviate
        loaded_tree = await Tree.import_from_weaviate(
            collection_name="Test_ELYSIA_save_load_collection_llm",
            conversation_id=tree.conversation_id,
            client_manager=client_manager,
        )

        assert tree.user_id == loaded_tree.user_id
        assert tree.conversation_id == loaded_tree.conversation_id
        assert (
            loaded_tree.tree_data.conversation_history
            == tree.tree_data.conversation_history
        )

        # now ask about his email without referring to the product name
        response, _ = loaded_tree("what about its price?")

        # check that the output includes emails from edward
        # Use a more appropriate metric for the test: check that the response references Edward's email.
        metric = metrics.GEval(
            name="Check the price of the product is referenced",
            criteria=(
                "The response should reference the price of the product. "
                "The answer should be about the price of the product, and should not hallucinate information if not present. "
                "The price of the product should be 39.99"
            ),
            evaluation_params=[
                LLMTestCaseParams.INPUT,
                LLMTestCaseParams.ACTUAL_OUTPUT,
                LLMTestCaseParams.EXPECTED_OUTPUT,
                LLMTestCaseParams.RETRIEVAL_CONTEXT,
            ],
        )

        test_case = LLMTestCase(
            input="what about its price?",
            actual_output=response,
            expected_output="""
            There should be a response about the Yoga Mat with Alignment Lines.
            Ensure that the response talks about the price of the product (39.99 - currency unknown, currency relevant to the test).
            You are only testing if the price of the product is referenced and correct.
            """,
            retrieval_context=[str(obj) for obj in objects],
        )

        res = evaluate(test_cases=[test_case], metrics=[metric])
        for test_case in res.test_results:
            assert test_case.success, test_case.metrics_data[0].reason

    finally:
        async with client_manager.connect_to_async_client() as client:
            if await client.collections.exists(collection_name):
                await client.collections.delete(collection_name)

            if await client.collections.exists("Test_ELYSIA_save_load_collection_llm"):
                await client.collections.delete("Test_ELYSIA_save_load_collection_llm")

        await client_manager.close_clients()



# Prompt Executors
#
import dspy
import dspy.predict

from elysia.util.elysia_chain_of_thought import ElysiaChainOfThought

# LLM
from elysia.objects import Response, Tool
from elysia.tools.text.objects import TextWithTitle, TextWithCitations
from elysia.tools.text.prompt_templates import (
    SummarizingPrompt,
    TextResponsePrompt,
    CitedSummarizingPrompt,
)

# Objects
from elysia.tree.objects import TreeData
from elysia.util.client import ClientManager


class CitedSummarizer(Tool):
    def __init__(self, **kwargs):
        super().__init__(
            name="cited_summarize",
            description="""
            Summarize retrieved information for the user when all relevant data has been gathered.
            Provides a text response, and may end the conversation, but unlike text_response tool, can be used mid-conversation.
            Avoid for general questions where text_response is available.
            Summarisation text is directly displayed to the user.
            Most of the time, you can choose end_actions to be True to end the conversation with a summary.
            This is a good way to end the conversation.
            """,
            status="Summarizing...",
            inputs={},
            end=True,
        )

    async def is_tool_available(
        self,
        tree_data: TreeData,
        base_lm: dspy.LM,
        complex_lm: dspy.LM,
        client_manager: ClientManager | None = None,
        **kwargs,
    ):
        return not tree_data.environment.is_empty()

    async def __call__(
        self,
        tree_data: TreeData,
        inputs: dict,
        base_lm: dspy.LM,
        complex_lm: dspy.LM,
        client_manager: ClientManager | None = None,
        **kwargs,
    ):
        summarizer = ElysiaChainOfThought(
            CitedSummarizingPrompt,
            tree_data=tree_data,
            reasoning=False,
            impossible=False,
            environment=True,
            tasks_completed=True,
            message_update=False,
        )

        summary = await summarizer.aforward(
            lm=base_lm,
        )

        yield TextWithCitations(
            cited_texts=summary.cited_text,
            title=summary.subtitle,
        )


class Summarizer(Tool):
    def __init__(self, **kwargs):
        super().__init__(
            name="summarize",
            description="""
            Summarize retrieved information for the user when all relevant data has been gathered.
            Provides a text response, and may end the conversation, but unlike text_response tool, can be used mid-conversation.
            Avoid for general questions where text_response is available.
            Summarisation text is directly displayed to the user.
            This summarizer will also include citations in the summary, so relevant information must be in the environment for this tool.
            Most of the time, you can choose end_actions to be True to end the conversation with a summary.
            This is a good way to end the conversation.
            """,
            status="Summarizing...",
            inputs={},
            end=True,
        )

    async def is_tool_available(
        self,
        tree_data: TreeData,
        base_lm: dspy.LM,
        complex_lm: dspy.LM,
        client_manager: ClientManager | None = None,
        **kwargs,
    ):
        # when the environment is non empty
        return not tree_data.environment.is_empty()

    async def __call__(
        self,
        tree_data: TreeData,
        inputs: dict,
        base_lm: dspy.LM,
        complex_lm: dspy.LM,
        client_manager: ClientManager | None = None,
        **kwargs,
    ):
        summarizer = ElysiaChainOfThought(
            SummarizingPrompt,
            tree_data=tree_data,
            environment=True,
            tasks_completed=True,
            message_update=False,
        )

        summary = await summarizer.aforward(
            lm=base_lm,
        )

        yield TextWithTitle(text=summary.summary, title=summary.subtitle)


class TextResponse(Tool):
    def __init__(self, **kwargs):
        super().__init__(
            name="final_text_response",
            description="",
            status="Writing response...",
            inputs={},
            end=True,
        )

    async def __call__(
        self,
        tree_data: TreeData,
        inputs: dict,
        base_lm: dspy.LM,
        complex_lm: dspy.LM,
        client_manager: ClientManager | None = None,
        **kwargs,
    ):
        text_response = ElysiaChainOfThought(
            TextResponsePrompt,
            tree_data=tree_data,
            environment=True,
            tasks_completed=True,
            message_update=False,
        )

        output = await text_response.aforward(
            lm=base_lm,
        )

        yield Response(text=output.response)


class FakeTextResponse(Tool):
    def __init__(self, **kwargs):
        super().__init__(
            name="text_response",
            description="""
            End the conversation. This should be used when the user has finished their query, or you have nothing more to do except reply.
            You should use this to answer conversational questions not related to other tools. But do not use this as a source of information.
            All information should be from the environment if answering a complex question or an explanation.
            If there is an error and you could not complete a task, use this tool to suggest a brief reason why.
            If, for example, there is a missing API key, then the user needs to add it to the settings (which you should inform them of).
            Or you cannot connect to weaviate, then the user needs to input their API keys in the settings.
            If there are no collections available, the user needs to analyze this in the 'data' tab.
            If there are other problems, and it looks like the user can fix it, then provide a suggestion.
            """,
            status="Writing response...",
            inputs={
                "text": {
                    "type": str,
                    "description": (
                        "The text to display to the user. Speak directly to them, do not say 'the user' or similar. "
                        "If you have achieved the goal, give a satisfying answer to their original prompt (user_prompt). "
                        "If you have not achieved the goal, explain. "
                        "If you know why, explain any shortcomings, maybe it was a limitation, a misunderstanding, a problem with the data. "
                        "Your suggestions should be useful to the user so they know if there is something possible to fix or do as a follow up. "
                        "Sometimes you can ask for clarification. "
                        "Be polite, professional, apologetic if necessary but above all - helpful! "
                        "Only provide a response in text, not using any variables or otherwise. "
                        "The text in this field is the DIRECT response to the user that will be displayed to them, and ending the conversation afterwards."
                    ),
                    "default": "",
                }
            },
            end=True,
        )

    async def __call__(
        self,
        tree_data: TreeData,
        inputs: dict,
        base_lm: dspy.LM,
        complex_lm: dspy.LM,
        client_manager: ClientManager | None = None,
        **kwargs,
    ):
        yield Response(text=inputs["text"])



from typing import List

import dspy

from elysia.tree.objects import Atlas
from elysia.tools.retrieval.util import (
    DataDisplay,
    QueryOutput,
)
from typing import Union


def construct_aggregation_output_prompt(vectorised: bool = False) -> str:
    output_prompt = """
    The AggregationOutput format contains these key components:

    1. target_collections: The collections you want to query (can be multiple, all will share the same aggregation query, but different queries can be chosen by providing multiple AggregationOutput objects)

    2. Filtering mechanism:
    - filter_buckets: Groups of filters connected with AND/OR logic
    - Each filter_bucket contains:
        - filters: List of property-specific filters or nested filter buckets
        - operator: "AND" or "OR" to connect filters within the bucket

    The FilterBucket is recursive - it can contain other FilterBuckets. This allows you to build complex logical expressions like (A AND B) OR (C AND D) by nesting buckets with different operators. For example:
    - A top-level bucket with OR operator containing:
        - A nested bucket with AND operator (for A AND B)
        - Another nested bucket with AND operator (for C AND D)
    - Example: For "January OR March", create a structure like: OR(AND(Jan-start, Jan-end), AND(Mar-start, Mar-end))

    3. Property filter types:
        - IntegerPropertyFilter: For integer comparisons (=, !=, <, >, <=, >=)
        - FloatPropertyFilter: For float comparisons (=, !=, <, >, <=, >=)
            Note that this is for float comparisons, not integer comparisons.
            If the property is an "int", you should use the IntegerPropertyFilter instead.
        - TextPropertyFilter: For text equality (=, !=) or pattern matching (LIKE). 
            When using LIKE, you can use the ? wildcard to match one unknown single character. e.g. "car?" matches "cart", "care", but not "car"
            You can also use the * wildcard to match any number of unknown characters. e.g. "car*" matches "car", "care", "carpet", etc
            "*car*" matches "car", "healthcare", etc
        - BooleanPropertyFilter: For boolean comparisons (=, !=)
        - DatePropertyFilter: For date comparisons (=, !=, <, >, <=, >=)
       Make sure to use the correct filter type for the property type. The property type is given in the schema (field.[field_name].type).
       If you use the incorrect filter type, the query will error. It is extremely important that you use the correct filter type!
       You CANNOT use any type of filter on an 'object' or 'object[]' property. It will raise an error!
       Note most filters are INCLUSIVE. You can use <=, <, >, >=, !=, to filter out results. But e.g. LIKE and = are INCLUSIVE.


    4. Aggregation operations (can choose many per AggregationOutput object, match the property type based on the field in the schema):
        - integer_property_aggregation: MIN, MAX, MEAN, MEDIAN, MODE, SUM on numeric fields
        - text_property_aggregation: TOP_OCCURRENCES for most frequent values
        - boolean_property_aggregation: TOTAL_TRUE, TOTAL_FALSE, PERCENTAGE_TRUE, PERCENTAGE_FALSE
        - date_property_aggregation: MIN, MAX, MEAN, MEDIAN, MODE on date fields
        Make sure to use the correct aggregation operation for the property type. The property type is given in the schema (field.[field_name].type).
        If you use the incorrect aggregation operation, the query will error. It is extremely important that you use the correct aggregation operation!

    5. Grouping:
        - groupby_property: Group aggregation results by a specific property value
        Grouping is optional but recommended for complex questions or those that require segmentation of results
    """

    if vectorised:
        output_prompt += """
    6. Optional Search:
        You can also use search to narrow the results for the aggregation.
        Use this sparingly. Only use if you judge it necessary based on the user's prompt. 
        - search_type: The type of search to use for the aggregation (hybrid, vector). Vector is semantic search, hybrid is a combination of semantic and keyword search. If None, no search is used.
        - search_query: The text to search for (using vector search). Only set this if search_type is specified.
        - limit: The number of results to narrow down to with the search. Only set this if search_type is specified.
            The limit should be large - aggregations are looking for statistics across the dataset.
            Judge this on the size of the dataset and the potential number of results from the search.
            Limit should be greater than 100 unless you judge the task requires a smaller limit.
    """

    return output_prompt


class AggregationPrompt(dspy.Signature):
    """
    You are a database query builder specializing in converting natural language questions into structured aggregation queries for Weaviate databases.
    You are part of an ensemble, a larger system of an agentic framework the user is interacting with, so you are not the only LLM they interact with.
    Given a user question, your job is to formulate precise aggregation queries based on the provided schema information.

    When constructing aggregation queries:
    1. Identify the relevant collection(s) from the schemas provided
    2. Choose the appropriate aggregation operation based on the user's intent (MIN, MAX, TOP_OCCURRENCES, etc.)
    3. Build filter_buckets when specific conditions are needed, using nested buckets for complex logic
    4. Apply grouping when needed to segment the aggregation results

    Provide detailed reasoning about your approach, explaining why you chose specific collections, properties, operations, and filters.
    In your reasoning also be detailed about the nested conditions of filter buckets.
    Include all of your thoughts in any reasoning fields, as this will be used to communicate to later models about your decision process here.

    The schema is provided in the `collection_schemas` field. This is important to use to understand the data types and properties of the collection.
    Ensure that you choose the correct filter types and aggregation operations for the properties in the schema, including values you filter/aggregate on, they must match the data type of the property.
        E.g., if the property is an integer, do not specify a float such as 1.0, it must be 1.
    Use the groups in the schema to understand the data for filters and aggregation operations.

    Return a list of AggregationOutput objects that satisfy the user's request. If multiple approaches are possible, return multiple queries.
    If the aggregation is impossible to execute, set impossible=True and return None.

    Remember that this is NOT the only query being performed in the larger program! So if you need to aggregate as a first or second step towards achieving a goal,
    you should do so, and explain in your reasoning what your future steps are towards this goal.
    You can only aggregate, which is creating summary statistics of data.
    If you think the goal is better suited to another task such as searching, you should still create an aggregation query to answer the question as best as you can.

    IMPORTANT NOTE: any query you perform will be on the FULL collection, not a filtered subset from any previous steps.
    Do not enclose any answers in ```python or ```.
    """

    available_collections: list[str] = dspy.InputField(
        description="A list of collections that are available to query. You can ONLY choose from these collections."
    )
    previous_aggregation_queries: list[str] = dspy.InputField(
        desc="""
        For each collection, a list of existing code that has been used to query the collection.
        This can be used to avoid generating duplicate queries.
        If any of the collections have an empty list, you are generating the first query for that collection.
        """.strip(),
        format=list[dict],
    )


def construct_query_output_prompt(vectorised: bool = False) -> str:
    output_prompt = f"""
    The QueryOutput format contains these key components:

    1. target_collection: The collection you want to query
    2. search_type: Choose from:
        - "filter_only": Standard SQL style search with filters, sorting, limits, fetching etc.
        - "keyword": Pure text-matching search
    """
    if vectorised:
        output_prompt += """
        - "vector": Semantic search using embeddings
        - "hybrid": Combined keyword and vector search
    """

    output_prompt += """
    3. search_query: The text to search for (optional, not used with filter_only)

    4. Filtering mechanism:
        - filter_buckets: Groups of filters connected with AND/OR logic
        - Each filter_bucket contains:
            - filter: List of property-specific filters or nested filter buckets
            - operator: "AND" or "OR" to connect filters within the bucket

        The FilterBucket is recursive - it can contain other FilterBuckets. This allows you to build complex logical expressions like (A AND B) OR (C AND D) by nesting buckets with different operators. For example:
        - A top-level bucket with OR operator containing:
            - A nested bucket with AND operator (for A AND B)
            - Another nested bucket with AND operator (for C AND D)
        - Example: For "January OR March", create a structure like: OR(AND(Jan-start, Jan-end), AND(Mar-start, Mar-end))

    5. Property filter types:
        - IntegerPropertyFilter: For integer comparisons (=, !=, <, >, <=, >=)
            Note that you cannot filter on integers with float values. 
            If you want to make a comparison between an integer property with a float value,
            you should convert the float value to an integer manually.
        - FloatPropertyFilter: For float comparisons (=, !=, <, >, <=, >=)
            Note that this is for float comparisons, not integer comparisons.
            If the property is an "int", you should use the IntegerPropertyFilter instead.
        - TextPropertyFilter: For text equality (=, !=) or pattern matching (LIKE), or IS_NULL
            When using LIKE, you can use the ? wildcard to match one unknown single character. e.g. "car?" matches "cart", "care", but not "car"
            You can also use the * wildcard to match any number of unknown characters. e.g. "car*" matches "car", "care", "carpet", etc
            "*car*" matches "car", "healthcare", etc
        - BooleanPropertyFilter: For boolean comparisons (=, !=, IS_NULL)
        - DatePropertyFilter: For date comparisons (=, !=, <, >, <=, >=, IS_NULL)
        - ListPropertyFilter: For list comparisons (=, !=, CONTAINS_ANY, CONTAINS_ALL, IS_NULL) (only if the property type is a list, e.g. "text[]", or "number[]")
        Make sure to use the correct filter type for the property type. The property type is given in the schema (field.[field_name].type).
        If you use the incorrect filter type, the query will error. It is extremely important that you use the correct filter type!
        You can also use the `length` parameter on NumberPropertyFilter/ListPropertyFilter ONLY (even if the property is not a number) to filter by the length of a list,
            ONLY if `isLengthIndexed` is set to True in the schema.
            (found in `index_properties.isLengthIndexed` in the schema)
        When `length` is set to True, you are not searching the properties, you are filtering based on the _length_ of the properties
            (if the property is a list, then this is the list length, if the property is a string, then this is the character length)
        Do NOT use the length filter if the property is not indexed for length. It will raise an error!
        You can also use the IS_NULL filter to filter for properties that are None or null (not an empty list or string etc.), ONLY if `isNullIndexed` is set to True in the schema.
            (found in `index_properties.isNullIndexed` in the schema)
        Do NOT use the IS_NULL filter if the property is not indexed for null. It will raise an error!
        You CANNOT use any type of filter on an 'object' or 'object[]' property. It will raise an error!
        Note most filters are INCLUSIVE. You can use <=, <, >, >=, !=, to filter out results. But e.g. LIKE and = are INCLUSIVE.


    6. CreationTimeFilter: For date comparisons for object creation time metadata (=, <, >, <=, >=)
        - This is a special filter that is used to filter by the creation time of an object.
        - You cannot filter individual properties, only the creation time of objects.
        - The value to filter on should be datetime formatted as a string, e.g. "2021-01-01T00:00:00Z"
        - IMPORTANT: This is only achievable if the collection schema has property `isTimestampIndexed` set to True.
            (found in `index_properties.isTimestampIndexed` in the schema)
            Do NOT use this filter if the property is not set to True. It will raise an error!

    7. Additional query parameters:
        - sort_by: Property and direction for sorting results
        - limit: Maximum number of results to return (default: 5)
        - groupby_property: Group results by a specific property value

    When constructing queries:
    1. Identify the relevant collection from the schemas provided
    2. Choose the appropriate search_type based on the user's intent
    3. Build filter_buckets when specific conditions are needed, using nested buckets for logic
    4. Apply sorting and limits as needed
    """

    if vectorised:
        output_prompt += """
    Tips for query_output:
    1.  a. Use 'hybrid' if you are unsure of the search type (but still need a search term), as this combines semantic and keyword search.
        b. Use 'filter_only' if there are no search terms needed.
        c. 'vector' and 'keyword' are good but should be reserved for specific cases where they are needed.
            e.g. keyword is good for looking for when the search terms WILL appear in the text, but do NOT use this as a filter if you can use a field instead
                    vector is good for matching meaning only, and the words themselves are not that important
    """
    else:
        output_prompt += """
    Tips for query_output:
    1.  a. Use 'filter_only' if there are no search terms needed.
        b. 'keyword' should be used as a more fine-grained way of reducing the result set down - it performs BM25 search on the text.
            e.g. keyword is good for looking for when the search terms WILL appear in the text, but do NOT use this as a filter if you can use fields for more direct filtering instead
    """

    output_prompt += """
    2. Err on the side of using larger limits, because you can parse the results later, and it's better to have more than too few.
        Additionally, if this query is part of a larger task, you may need to retrieve a lot of objects to infer correlations or trends.
        If you are looking for overall patterns or trends, use a larger limit (greater than 10).
    3. Do not use a search term to look for specific categories of items if there is a corresponding field in the schema you can use instead as a filter.
    4. You cannot combine sorting with a search term (mutually exclusive), so if sorting is important, do not specify a search term.
    5. Nesting filter buckets is really powerful, and if the queries are particularly complex, definitely use this feature.
    6. Do not be afraid of using lots of features (filters, limits, search terms) at once, as this may be required for bigger queries.
    7. Do not feel the need to answer everything all at once. You are just one step in a larger program. This step can be repeated at a later date.
    8. If you are querying multiple collections for the same thing, do not repeat the same query for each collection, just provide multiple collections a single query_output.
    9. Always make sure your filter matches the property type in the dataset.

    Return None if query is impossible to execute.
    """
    return output_prompt


class QueryCreatorPrompt(dspy.Signature):
    """
    You are a database query builder. Your task is to convert natural language questions into structured queries using the provided schema information.

    Instructions:
    1. Carefully read the user's question and the schema details.
    2. Formulate clear, precise queries that best address the user's request.
    3. If the user's question is missing important search terms, use your knowledge to identify and include what should be searched, even if not explicitly mentioned.
    4. You may create multiple queries for the same collection if needed (e.g., with different search terms or filters). Clearly explain your reasoning for each.
    5. Write all your reasoning in the provided reasoning fields. This will help later models understand your decision process.
    6. Return a list of QueryOutput objects that fully satisfy the user's request. If there are multiple valid approaches, include multiple queries.
    7. Remember: Each query operates on the entire collection, not on a subset filtered by previous steps. Ensure filters and arguments are consistent across all queries.
    8. Do not write any code or define new classes—just provide the correct arguments as specified by the schema.

    The schema is provided in the `collection_schemas` field. This is important to use to understand the data types and properties of the collection.
    Ensure that you choose the correct filter types for the properties in the schema, including values you filter on, they must match the data type of the property.
        E.g., if the property is an integer, do not specify a float such as 1.0, it must be 1.
    Use the groups in the schema to understand the data for filters.

    Be concise, clear, and thorough in your reasoning and query construction.
    """

    advanced_reasoning: str = dspy.OutputField(
        desc="""
        In draft form (no more than 5 words per sentence), provide some extra reasoning and thinking.
        Provide one sentence for each of the following:
            - Whether the collection isNullIndexed, isLengthIndexed, or isTimestampIndexed
                If so, do you need to use `length`, `IS_NULL`, or number_property filters?
            - How the filter buckets are nested, and why you chose this structure.
                This logic is important, what should be an OR and what should be an AND?
            - What type of search to perform, and why you chose this search type.
            - If there are any fields to search on, why you chose these fields.
            - How you are defining the query_output, and that you are providing only the schema, no extra code.
              This includes defining any variables, you cannot do this. EVERYTHING MUST BE IN THE QUERY_OUTPUT (WITHOUT EXTRA DEFINITIONS).
              NO CODE!
        """.strip()
    )
    available_collections: list[str] = dspy.InputField(
        description="A list of collections that are available to query. You can ONLY choose from these collections."
    )
    previous_queries: list[str] = dspy.InputField(
        desc="""
        For each collection, a list of existing code that has been used to query the collection.
        This can be used to avoid generating duplicate queries.
        If any of the collections have an empty list, you are generating the first query for that collection.
        """.strip(),
        format=list[dict],
    )
    collection_display_types: dict[str, list[str]] = dspy.InputField(
        desc="""
        A dictionary of the display types for each collection. Use this to determine the display type in `display_type`, for each collection.
        The output of `display_type` must be one of the values in the list for the dictionary entry for that collection.
        """.strip(),
    )
    display_type_descriptions: dict[str, str] = dspy.InputField(
        desc="""
        Descriptions of each of the available display types.
        Use this to determine your output display type(s).
        The keys are the display types, and the values are the descriptions.
        """.strip(),
        format=dict[str, str],
    )
    searchable_fields: dict[str, list[str]] = dspy.InputField(
        desc="""
        A dictionary of the named vector fields to search for each collection. Use this to determine the field in `vector_to_search`, for each collection.
        IMPORTANT: these are not property names, they are the names of vectors to search on (if using hybrid or vector search).
        The keys are the collection names, and the values are the fields that are possible to perform the search on.
        """.strip(),
        format=dict[str, list[str]],
    )

    fields_to_search: dict[str, list[str] | None] = dspy.OutputField(
        desc="""
        A dictionary of the named vector fields to search for each collection.
        IMPORTANT: these are not property names, they are the names of vectors to search on (if using hybrid or vector search).
        Always only choose a field from the `searchable_fields` list.
        The keys are the collection names, and the values are the fields to search for.
        These come from the `searchable_fields` list. If this list is empty, the value should be None.
        Otherwise, you should choose the field(s) that would be most relevant to the user's query.
        If in doubt, set the value inside the dictionary to None.
        """.strip()
    )
    data_display: dict[str, DataDisplay] = dspy.OutputField(
        desc="""
        Some metadata about the data you will be returning.
        The keys are the collection names, and the values are the DataDisplay objects.
        The `display_type` should match ONE of the display types in the `collection_display_types` for the corresponding collection.
        The `summarise_items` should be True if the user has specifically asked for summaries of the objects (in an itemised fashion, not an overall summary).
        Otherwise, it should be False.
        """.strip()
    )



from typing import AsyncGenerator, Union
from logging import Logger
from pydantic import BaseModel, Field

from rich import print
from rich.panel import Panel

import dspy

from elysia.util.elysia_chain_of_thought import ElysiaChainOfThought
from elysia.objects import Response, Status, Tool, Error, Result, Return, Retrieval
from elysia.tools.retrieval.chunk import AsyncCollectionChunker
from elysia.tools.retrieval.objects import (
    ConversationRetrieval,
    DocumentRetrieval,
    MessageRetrieval,
)
from elysia.tools.retrieval.prompt_templates import (
    QueryCreatorPrompt,
    construct_query_output_prompt,
)
from elysia.tools.retrieval.util import (
    execute_weaviate_query,
    QueryError,
    QueryOutput,
    NonVectorisedQueryOutput,
)
from elysia.tree.objects import TreeData
from elysia.util.client import ClientManager
from elysia.util.objects import TrainingUpdate, TreeUpdate, FewShotExamples
from elysia.util.return_types import all_return_types


class Query(Tool):
    """
    Elysia Query Tool
    Queries the knowledge base with Weaviate using hybrid, semantic, or keyword search, applying filters and more.
    These are dynamically decided by an LLM, based on the user prompt and the collections available.

    Additionally, this Tool can perform the following actions:
    - Chunk collections that are too large to vectorise effectively
    - Dynamically return the retrieved objects to the frontend according to the display type decided by the LLM
    - Set a flag to summarise retrieved objects in the tree

    """

    def __init__(
        self,
        logger: Logger | None = None,
        summariser_in_tree: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(
            name="query",
            description="""
            Retrieves and displays specific data entries from the collections.
            Then, query with semantic search, keyword search, or a combination of both.
            Queries can be filtered, sorted, and more.
            Retrieving and displaying specific data entries rather than performing calculations or summaries.
            Do not use 'query' as a preliminary filtering step when 'aggregate' can achieve the same result more efficiently (if 'aggregate' is available).
            """,
            status="Querying...",
            inputs={
                "collection_names": {
                    "description": "the names of the collections that are most relevant to the query. If you are unsure, give all collections.",
                    "type": list[str],
                    "default": [],
                },
            },
            end=False,
        )

        self.logger = logger
        self.summariser_in_tree = summariser_in_tree
        self.retrieval_map = {
            "conversation": ConversationRetrieval,
            "message": MessageRetrieval,
            "document": DocumentRetrieval,
        }

    async def is_tool_available(
        self,
        tree_data: TreeData,
        base_lm: dspy.LM,
        complex_lm: dspy.LM,
        client_manager: ClientManager,
    ) -> bool:
        """
        Only available if:
        1. There is a Weaviate connection
        2. There are collections available
        If this tool is not available, inform the user that they should make sure they have set the WCD_URL and WCD_API_KEY in the settings.
        And also they should make sure they have added collections to the tree.
        """
        return client_manager.is_client and tree_data.collection_names != []

    def _find_previous_queries(
        self, environment: dict, collection_names: list[str]
    ) -> dict:
        previous_queries = {}
        # create a dict of collection_name: previous_queries
        for collection_name in collection_names:
            # check if a query has happened at all for this collection
            if "query" in environment and collection_name in environment["query"]:
                for query in environment["query"][collection_name]:
                    if "query_output" in query["metadata"]:
                        if collection_name in previous_queries:
                            previous_queries[collection_name].append(
                                query["metadata"]["query_output"]
                            )
                        else:
                            previous_queries[collection_name] = [
                                query["metadata"]["query_output"]
                            ]

        return previous_queries

    def _evaluate_content_field(
        self, metadata_fields: list[dict]
    ) -> tuple[str | None, float | None]:
        """
        Find the largest text field in the metadata.
        Returns the field name with the highest mean value among text fields.
        """
        # Filter for text fields only
        text_fields = {
            field["name"]: field
            for field in metadata_fields
            if (field.get("type") == "text" and field.get("mean") is not None)
        }

        if not text_fields:
            return None, None

        # Find the text field with the largest mean value
        max_field = max(text_fields, key=lambda x: text_fields[x].get("mean", 0))
        return max_field, text_fields[max_field]["mean"]

    def _evaluate_needs_chunking(
        self,
        display_type: str,
        query_type: str,
        schema: dict,
        threshold: int = 400,  # number of tokens to be considered large enough to chunk
    ) -> bool:
        content_field, content_len = self._evaluate_content_field(
            schema["fields"],
        )

        return (
            content_field is not None
            and content_len > threshold
            and query_type != "filter_only"
            and display_type
            == "document"  # for the moment, the only type we chunk is document
        )

    def _fix_collection_names(
        self, collection_names: list[str], schemas: dict
    ) -> list[str]:
        return [
            correct_collection_name
            for correct_collection_name in list(schemas.keys())
            if correct_collection_name.lower()
            in [collection_name.lower() for collection_name in collection_names]
        ]

    def _fix_collection_names_in_dict(
        self, collection_dict: dict, schemas: dict
    ) -> dict:
        true_collection_names = list(schemas.keys())
        truth_map = {k.lower(): k for k in true_collection_names}
        return {truth_map.get(k.lower(), k): v for k, v in collection_dict.items()}

    def _parse_weaviate_error(self, error_message: str) -> str:
        if "no api key found" in error_message.lower():
            api_error_message = error_message[
                (
                    error_message.find("environment variable under ")
                    + len("environment variable under ")
                ) :
            ]
            api_error_message = api_error_message[: api_error_message.find('"\n')]
            return (
                f"The following API key is needed: {api_error_message} for this collection. "
                "The user must set the API key in the settings. "
                "Alternatively, you can try keyword (BM25) search instead."
            )

        return error_message

    async def __call__(
        self,
        tree_data: TreeData,
        inputs: dict,
        base_lm: dspy.LM,
        complex_lm: dspy.LM,
        client_manager: ClientManager,
        **kwargs,
    ) -> AsyncGenerator[Return | TreeUpdate | Error | TrainingUpdate, None]:
        """
        Can perform three main functions:
        1. Query the knowledge base with Weaviate using hybrid, semantic, or keyword search, applying filters and more.
        2. Automatically chunk collections that are too large to vectorise effectively
        3. Summarise retrieved objects if requested
        Chunking creates a parallel quantized collection of chunks with cross references to the original collection.
        When filtering on the parallel chunked collection, filters are applied to the original collection by cross referencing.
        """

        if self.logger:
            self.logger.debug("Query Tool called!")
            self.logger.debug(f"Inputs: {inputs}")

        # Get the collection schemas (truth)
        schemas = tree_data.output_collection_metadata(with_mappings=True)

        # Find matching collection names (to match case sensitivity)
        if inputs["collection_names"] == [] or inputs["collection_names"] is None:
            collection_names = list(schemas.keys())
        else:
            collection_names = self._fix_collection_names(
                inputs["collection_names"], schemas
            )

        # Subset the schemas to just the collection names
        schemas = {
            collection_name: schemas[collection_name]
            for collection_name in collection_names
        }
        if self.logger:
            self.logger.debug(f"Collection names received: {collection_names}")

        # Set up the dspy model
        query_creator_prompt = QueryCreatorPrompt

        # check if collections are vectorised
        vectorised = any(
            (
                schemas[collection_name]["vectorizer"] is not None
                or (
                    schemas[collection_name]["named_vectors"] is not None
                    and schemas[collection_name]["named_vectors"] != []
                )
            )
            for collection_name in collection_names
        )

        if vectorised:
            query_creator_prompt = query_creator_prompt.append(
                name="query_outputs",
                type_=Union[list[QueryOutput], None],
                field=dspy.OutputField(
                    desc=construct_query_output_prompt(True),
                ),
            )
        else:
            query_creator_prompt = query_creator_prompt.append(
                name="query_outputs",
                type_=Union[list[NonVectorisedQueryOutput], None],
                field=dspy.OutputField(
                    desc=construct_query_output_prompt(False),
                ),
            )

        query_generator = ElysiaChainOfThought(
            query_creator_prompt,
            tree_data=tree_data,
            environment=True,
            collection_schemas=True,
            tasks_completed=True,
            message_update=True,
            collection_names=collection_names,
        )

        # Get some metadata about the collection
        previous_queries = self._find_previous_queries(
            tree_data.environment.environment,
            tree_data.collection_data.collection_names,
        )

        if self.logger:
            self.logger.debug(f"Previous queries: {previous_queries}")

        # get display types
        display_types = tree_data.output_collection_return_types()
        display_types = {
            collection_name: display_types[collection_name]
            for collection_name in collection_names
        }  # subset to just our collection names

        # get display type descriptions
        display_type_descriptions = {
            display_type: all_return_types[display_type]
            for collection_name in collection_names
            for display_type in display_types[collection_name]
        }

        # get searchable fields
        searchable_fields = {
            collection_name: [
                named_vector["name"]
                for named_vector in schemas[collection_name]["named_vectors"]
                if named_vector["enabled"]
            ]
            for collection_name in collection_names
            if schemas[collection_name]["named_vectors"] is not None
        }

        # Generate query with LLM
        try:
            if tree_data.settings.USE_FEEDBACK:
                query, example_uuids = (
                    await query_generator.aforward_with_feedback_examples(
                        feedback_model="query",
                        client_manager=client_manager,
                        base_lm=base_lm,
                        complex_lm=complex_lm,
                        available_collections=collection_names,
                        previous_queries=previous_queries,
                        collection_display_types=display_types,
                        display_type_descriptions=display_type_descriptions,
                        searchable_fields=searchable_fields,
                        num_base_lm_examples=6,
                        return_example_uuids=True,
                    )
                )
            else:
                query = await query_generator.aforward(
                    lm=complex_lm,
                    available_collections=collection_names,
                    previous_queries=previous_queries,
                    collection_display_types=display_types,
                    display_type_descriptions=display_type_descriptions,
                    searchable_fields=searchable_fields,
                )

        except Exception as e:
            yield Error(error_message=str(e))
            return

        if self.logger and query.query_outputs is not None:
            self.logger.debug(f"Query: {query.query_outputs}")
            self.logger.debug(f"Fields to search: {query.fields_to_search}")
            self.logger.debug(f"Data display: {query.data_display}")

        # Yield results to front end
        yield Response(text=query.message_update)
        if tree_data.settings.USE_FEEDBACK:
            yield FewShotExamples(uuids=example_uuids)

        # Return if model deems query impossible
        if (
            query.impossible
            or query.query_outputs is None
            or all(q is None for q in query.query_outputs)
            or len(query.query_outputs) == 0
        ):

            for collection_name in collection_names:
                metadata = {
                    "collection_name": collection_name,
                    "impossible": tree_data.user_prompt,
                    "impossible_reasoning": (
                        query.reasoning
                        if tree_data.settings.COMPLEX_USE_REASONING
                        else ""
                    ),
                    "query_output": (
                        query.query_outputs.model_dump()
                        if query.query_outputs is not None
                        else None
                    ),
                }
                yield Retrieval([], metadata)

            if self.logger:
                self.logger.warning(
                    f"Model judged query to be impossible. Returning to the decision tree..."
                )
            return

        # extract and error handle the query output
        for current_query_output in query.query_outputs:

            current_query_output.target_collections = self._fix_collection_names(
                current_query_output.target_collections, schemas
            )

            for collection_name in current_query_output.target_collections:
                if collection_name not in collection_names:
                    yield Error(
                        feedback=(
                            f"Collection {collection_name} in target_collections field of the query output, "
                            "but not found in the available collections. "
                            "Make sure you are using the correct collection names."
                        ),
                    )
                    return

        query.data_display = self._fix_collection_names_in_dict(
            query.data_display, schemas
        )
        query.fields_to_search = self._fix_collection_names_in_dict(
            query.fields_to_search, schemas
        )

        for collection_name in query.data_display:

            if collection_name not in collection_names:
                yield Error(
                    feedback=(
                        f"Collection {collection_name} in data_display keys, "
                        "but not found in the available collections. "
                        "Make sure you are using the correct collection names."
                    ),
                )
                return

        for collection_name in query.fields_to_search:
            if collection_name not in collection_names:
                yield Error(
                    feedback=(
                        f"Collection {collection_name} in fields_to_search keys, "
                        "but not found in the available collections. "
                        "Make sure you are using the correct collection names."
                    ),
                )
                return

        # if the fields_to_search is non-empty but there are no named vectors, ignore this and reset it
        for collection_name in query.fields_to_search:
            if query.fields_to_search[collection_name] is not None and (
                schemas[collection_name]["named_vectors"] is None
                or searchable_fields[collection_name] == []
            ):
                query.fields_to_search[collection_name] = None

        # Go through outputs for each unique query
        for i, query_output in enumerate(query.query_outputs):
            collection_names = query_output.target_collections.copy()

            for collection_name in collection_names:

                display_type = query.data_display[collection_name].display_type

                # Evaluate if this collection/query needs chunking
                needs_chunking = self._evaluate_needs_chunking(
                    display_type,
                    query_output.search_type,
                    schemas[collection_name],
                )
                content_field, _ = self._evaluate_content_field(
                    schemas[collection_name]["fields"],
                )  # used for chunking

                if needs_chunking and content_field is not None:

                    if self.logger:
                        self.logger.debug(f"Chunking {collection_name}")

                    # set up chunking (create reference in this collection)
                    collection_chunker = AsyncCollectionChunker(collection_name)
                    await collection_chunker.create_chunked_reference(
                        content_field, client_manager
                    )

                    # create a copy of the query output
                    query_output_copy = query_output.model_copy(deep=True)

                    # update the limit to be larger
                    query_output_copy.limit = query_output.limit * 3

                    # update the collection to be the unchunked collection ONLY
                    query_output_copy.target_collections = [collection_name]

                    # run this augmented query for the unchunked collection
                    async with client_manager.connect_to_async_client() as client:
                        try:
                            unchunked_response, _ = await execute_weaviate_query(
                                client,
                                query_output_copy,
                                reference_property="isChunked",
                                named_vector_fields=query.fields_to_search,
                                property_types={
                                    collection_name: {field["name"]: field["type"]}
                                    for field in schemas[collection_name]["fields"]
                                },
                                schema=schemas,
                            )
                        except QueryError as e:
                            yield Error(feedback=str(e))
                            continue
                        except Exception as e:
                            yield Error(error_message=str(e))
                            continue

                    # chunk the unchunked response
                    yield Status(
                        f"Chunking {len(unchunked_response[0].objects)} objects in {collection_name}..."
                    )
                    await collection_chunker(
                        unchunked_response[0], content_field, client_manager
                    )

                    # modify the original query output to be the chunked collection
                    for j, c in enumerate(query_output.target_collections):
                        if c == collection_name:
                            query_output.target_collections[j] = (
                                f"ELYSIA_CHUNKED_{collection_name.lower()}__"
                            )
                    query.fields_to_search[
                        f"ELYSIA_CHUNKED_{collection_name.lower()}__"
                    ] = None

                    if self.logger:
                        self.logger.debug(
                            f"Chunked {collection_name} and updated query output to {query_output.target_collections}"
                        )

            # Execute query within Weaviate
            async with client_manager.connect_to_async_client() as client:
                try:
                    responses, code_strings = await execute_weaviate_query(
                        client,
                        query_output,
                        named_vector_fields=query.fields_to_search,
                        property_types={
                            collection_name: {
                                field["name"]: field["type"]
                                for field in schemas[collection_name]["fields"]
                            }
                            for collection_name in collection_names
                        },
                        schema=schemas,
                    )
                except QueryError as e:
                    yield Error(feedback=str(e))
                    continue
                except Exception as e:
                    for collection_name in collection_names:
                        yield Error(error_message=str(e))
                    continue

                yield Status(
                    f"Retrieved {sum(len(x.objects) for x in responses)} objects from {len(collection_names)} collections..."
                )

            for k, (collection_name, response) in enumerate(
                zip(collection_names, responses)
            ):
                if self.logger and self.logger.level <= 20:
                    print(
                        Panel.fit(
                            code_strings[k],
                            title=f"{collection_name} (Weaviate Query)",
                            border_style="yellow",
                            padding=(1, 1),
                        )
                    )

                # Get display type and summarise items bool
                display_type = query.data_display[collection_name].display_type
                summarise_items = query.data_display[collection_name].summarise_items

                # get query output formatted
                query_output_formatted = query_output.model_dump()
                query_output_formatted["target_collections"] = collection_names

                # Get the objects from the response
                objects = []
                for obj in response.objects:
                    objects.append({k: v for k, v in obj.properties.items()})
                    objects[-1]["uuid"] = str(obj.uuid)

                # Write various metadata for LLM parsing
                metadata = {
                    "collection_name": collection_name,
                    "display_type": display_type,
                    "needs_summarising": summarise_items,
                    "query_text": query_output.search_query,
                    "query_type": query_output.search_type,
                    "chunked": self._evaluate_needs_chunking(
                        display_type,
                        query_output.search_type,
                        schemas[collection_name],
                    ),
                    "query_output": query_output_formatted,
                    "code": {
                        "language": "python",
                        "title": "Query",
                        "text": code_strings[k],
                    },
                }

                # Create the retrieval object
                if display_type in self.retrieval_map:
                    output = self.retrieval_map[display_type](
                        objects,
                        metadata,
                        mapping=(
                            schemas[collection_name]["mappings"][display_type]
                            if display_type != "table"
                            else None
                        ),
                    )
                else:
                    output = Retrieval(
                        objects,
                        metadata,
                        payload_type=display_type,
                        name=collection_name,
                        mapping=(
                            schemas[collection_name]["mappings"][display_type]
                            if display_type != "table"
                            else None
                        ),
                    )

                # If the display type is document or conversation, initialise the object (requires some async operations)
                if isinstance(output, DocumentRetrieval) or isinstance(
                    output, ConversationRetrieval
                ):
                    await output.async_init(client_manager)

                if summarise_items and self.summariser_in_tree:
                    if (
                        "items_to_summarise"
                        not in tree_data.environment.hidden_environment
                    ):
                        tree_data.environment.hidden_environment[
                            "items_to_summarise"
                        ] = []
                    tree_data.environment.hidden_environment[
                        "items_to_summarise"
                    ].append(output)
                else:
                    yield output

        # if successful, yield training update
        yield TrainingUpdate(
            module_name="query",
            inputs={
                "available_collections": collection_names,
                "previous_queries": previous_queries,
                "collection_display_types": display_types,
                "searchable_fields": searchable_fields,
                **tree_data.to_json(),
            },
            outputs=query.__dict__["_store"],
        )
        if self.logger:
            self.logger.debug("Query Tool finished!")



from fastapi import APIRouter, Depends
from fastapi.responses import JSONResponse

from elysia.api.dependencies.common import get_user_manager
from elysia.api.services.user import UserManager

# Logging
from elysia.api.core.log import logger

router = APIRouter()


@router.get("/{user_id}/saved_trees")
async def get_saved_trees(
    user_id: str,
    user_manager: UserManager = Depends(get_user_manager),
):

    headers = {"Cache-Control": "no-cache"}

    user = await user_manager.get_user_local(user_id)
    save_location_client_manager = user["frontend_config"].save_location_client_manager
    if not save_location_client_manager.is_client:
        logger.warning(
            "In /get_saved_trees API, "
            "no valid destination for trees location found. "
            "Returning no error but an empty list of trees."
        )
        return JSONResponse(
            content={"trees": {}, "error": ""},
            status_code=200,
            headers=headers,
        )

    try:
        trees = await user_manager.get_saved_trees(
            user_id, save_location_client_manager
        )
        return JSONResponse(
            content={"trees": trees, "error": ""}, status_code=200, headers=headers
        )

    except Exception as e:
        logger.error(f"Error getting saved trees: {str(e)}")
        return JSONResponse(
            content={"trees": {}, "error": str(e)}, status_code=500, headers=headers
        )


@router.get("/{user_id}/load_tree/{conversation_id}")
async def load_tree(
    user_id: str,
    conversation_id: str,
    user_manager: UserManager = Depends(get_user_manager),
):

    headers = {"Cache-Control": "no-cache"}

    try:
        frontend_rebuild = await user_manager.load_tree(user_id, conversation_id)
        return JSONResponse(
            content={"rebuild": frontend_rebuild, "error": ""},
            status_code=200,
            headers=headers,
        )
    except Exception as e:
        logger.error(f"Error loading tree: {str(e)}")
        return JSONResponse(
            content={"rebuild": [], "error": str(e)},
            status_code=500,
            headers=headers,
        )


@router.post("/{user_id}/save_tree/{conversation_id}")
async def save_tree(
    user_id: str,
    conversation_id: str,
    user_manager: UserManager = Depends(get_user_manager),
):
    try:
        await user_manager.save_tree(user_id, conversation_id)
        return JSONResponse(content={"error": ""}, status_code=200)
    except Exception as e:
        logger.error(f"Error saving tree: {str(e)}")
        return JSONResponse(content={"error": str(e)}, status_code=500)


@router.delete("/{user_id}/delete_tree/{conversation_id}")
async def delete_tree(
    user_id: str,
    conversation_id: str,
    user_manager: UserManager = Depends(get_user_manager),
):
    try:
        await user_manager.delete_tree(user_id, conversation_id)
        return JSONResponse(content={"error": ""}, status_code=200)
    except Exception as e:
        logger.error(f"Error deleting tree: {str(e)}")
        return JSONResponse(content={"error": str(e)}, status_code=500)



import asyncio
import os
import pytest
from typing import Any
from weaviate.classes.query import Filter, QueryReference

from elysia.tree.objects import Environment
from elysia.objects import (
    Completed,
    Response,
    Status,
    Text,
    Update,
    Warning,
    Result,
)
from elysia.tools.retrieval.chunk import AsyncCollectionChunker
from elysia.tools.retrieval.objects import (
    ConversationRetrieval,
    DocumentRetrieval,
    MessageRetrieval,
)
from elysia.util.client import ClientManager
from elysia.util.parsing import format_dict_to_serialisable


@pytest.mark.asyncio
async def test_basic_result():
    result = Result(
        objects=[{"a": 1, "b": 2, "c": 3}],
        metadata={"metadata_a": 1, "metadata_b": 2, "metadata_c": 3},
        payload_type="test_type",
    )
    result_json = result.to_json()

    assert result_json == [{"a": 1, "b": 2, "c": 3}], "Objects should be equal"

    frontend_result = await result.to_frontend("user_id", "conversation_id", "query_id")

    # all id vars should be present
    for key in ["user_id", "conversation_id", "query_id"]:
        assert key in frontend_result, f"'{key}' not found in the frontend result"

    # type and payload type should be correct
    assert frontend_result["type"] == "result"
    assert frontend_result["payload"]["type"] == "test_type"

    # objects should be correct (but there can be more objects in frontend result)
    for i, obj in enumerate(result_json):
        for key in obj:
            assert (
                key in frontend_result["payload"]["objects"][i]
            ), f"'{key}' not found in the frontend result"

    # metadata should be the same
    assert frontend_result["payload"]["metadata"] == result.metadata


@pytest.mark.asyncio
async def test_result_with_mapping():
    objects = [{"a": 1, "b": 2, "c": 3, "d": 4, "e": 5, "f": 6}]
    mapping = {"new_a": "a", "new_b": "b", "new_c": "c"}
    unmapped_keys = ["d", "e", "f"]

    result = Result(
        objects=objects, metadata={}, mapping=mapping, unmapped_keys=unmapped_keys
    )
    result_json = result.to_json(True)

    for i, obj in enumerate(objects):
        for mapping_key in mapping:
            assert mapping_key in result_json[i]

        for obj_key in obj:
            if obj_key not in unmapped_keys:
                assert obj_key not in result_json[i]

        for unmapped_key in unmapped_keys:
            assert unmapped_key in result_json[i]

    frontend_result = await result.to_frontend("user_id", "conversation_id", "query_id")

    for i, obj in enumerate(objects):
        for mapping_key in mapping:
            assert mapping_key in frontend_result["payload"]["objects"][i]


def test_environment():
    environment = Environment()

    assert environment.is_empty()

    res = Result(
        objects=[{"a": 1, "b": 2}],
        metadata={"metadata_a": "hello", "metadata_b": 10},
        name="test_result",
    )
    environment.add("test_tool", res)

    assert not environment.is_empty()
    assert (
        environment.environment["test_tool"]["test_result"][0]["objects"][0]["a"] == 1
    )
    assert (
        environment.environment["test_tool"]["test_result"][0]["objects"][0]["b"] == 2
    )

    assert environment.environment["test_tool"]["test_result"][0]["metadata"] == {
        "metadata_a": "hello",
        "metadata_b": 10,
    }

    found_item = environment.find("test_tool", "test_result")
    assert found_item[0]["objects"][0]["a"] == 1
    assert found_item[0]["objects"][0]["b"] == 2
    assert found_item[0]["metadata"] == {
        "metadata_a": "hello",
        "metadata_b": 10,
    }

    found_item_index = environment.find("test_tool", "test_result", 0)
    assert found_item_index["objects"][0]["a"] == 1
    assert found_item_index["objects"][0]["b"] == 2
    assert found_item_index["metadata"] == {
        "metadata_a": "hello",
        "metadata_b": 10,
    }

    environment.replace("test_tool", "test_result", [{"c": 3, "d": 4}])
    assert (
        environment.environment["test_tool"]["test_result"][0]["objects"][0]["c"] == 3
    )
    assert (
        environment.environment["test_tool"]["test_result"][0]["objects"][0]["d"] == 4
    )

    environment.replace("test_tool", "test_result", [{"e": 5, "f": "hello"}], index=0)
    assert (
        environment.environment["test_tool"]["test_result"][0]["objects"][0]["e"] == 5
    )
    assert (
        environment.environment["test_tool"]["test_result"][0]["objects"][0]["f"]
        == "hello"
    )

    json_obj = environment.environment
    assert "test_tool" in json_obj
    assert "test_result" in json_obj["test_tool"]
    assert json_obj["test_tool"]["test_result"][0]["objects"][0]["e"] == 5
    assert json_obj["test_tool"]["test_result"][0]["objects"][0]["f"] == "hello"

    environment.remove("test_tool", "test_result")
    assert environment.is_empty()

    environment.add_objects(
        "test_tool2",
        "test_result2",
        [{"a": 1, "b": 2}],
        metadata={"item_number": "first"},
    )
    assert "test_tool2" in environment.environment
    assert "test_result2" in environment.environment["test_tool2"]
    assert (
        environment.environment["test_tool2"]["test_result2"][0]["objects"][0]["a"] == 1
    )
    assert (
        environment.environment["test_tool2"]["test_result2"][0]["objects"][0]["b"] == 2
    )
    assert (
        environment.environment["test_tool2"]["test_result2"][0]["metadata"][
            "item_number"
        ]
        == "first"
    )

    environment.add_objects(
        "test_tool2",
        "test_result2",
        [{"c": 1, "d": 2}],
        metadata={"item_number": "second"},
    )
    assert (
        environment.environment["test_tool2"]["test_result2"][1]["objects"][0]["c"] == 1
    )
    assert (
        environment.environment["test_tool2"]["test_result2"][1]["objects"][0]["d"] == 2
    )
    assert (
        environment.environment["test_tool2"]["test_result2"][1]["metadata"][
            "item_number"
        ]
        == "second"
    )

    assert (
        environment.environment["test_tool2"]["test_result2"][0]["objects"][0]["a"] == 1
    )
    assert (
        environment.environment["test_tool2"]["test_result2"][0]["objects"][0]["b"] == 2
    )

    environment.remove("test_tool2", "test_result2", 1)
    assert environment.environment["test_tool2"]["test_result2"] != []
    assert (
        environment.environment["test_tool2"]["test_result2"][0]["objects"][0]["a"] == 1
    )
    assert (
        environment.environment["test_tool2"]["test_result2"][0]["objects"][0]["b"] == 2
    )
    assert (
        environment.environment["test_tool2"]["test_result2"][0]["metadata"][
            "item_number"
        ]
        == "first"
    )
    assert len(environment.environment["test_tool2"]["test_result2"]) == 1


@pytest.mark.asyncio
async def test_updates():
    types = [
        Status,
        Warning,
    ]
    names = [
        "status",
        "warning",
        "error",
    ]
    for object_type, object_name in zip(types, names):
        result = object_type(
            "this is a test",
        )
        frontend_result = await result.to_frontend(
            "user_id", "conversation_id", "query_id"
        )
        assert frontend_result["type"] == object_name
        assert isinstance(frontend_result["payload"], dict)



import asyncio
import inspect
import spacy

from weaviate.classes.config import Configure, DataType, Property, ReferenceProperty
from weaviate.collections.classes.data import DataObject, DataReference
from weaviate.collections import CollectionAsync
from weaviate.collections.classes.internal import Object, QueryReturn
from weaviate.collections.classes.config_vectors import _VectorConfigCreate
from weaviate.exceptions import WeaviateInvalidInputError
from weaviate.util import generate_uuid5
from weaviate.client import WeaviateAsyncClient

from elysia.util.client import ClientManager
from elysia.util.collection import (
    async_get_collection_weaviate_data_types,
)


def chunked_collection_exists(
    collection_name: str, client_manager: ClientManager | None = None
):
    if client_manager is None:
        client_manager = ClientManager()

    with client_manager.connect_to_client() as client:
        return client.collections.exists(f"ELYSIA_CHUNKED_{collection_name.lower()}__")


def delete_chunked_collection(
    collection_name: str, client_manager: ClientManager | None = None
):
    if client_manager is None:
        client_manager = ClientManager()

    with client_manager.connect_to_client() as client:
        client.collections.delete(f"ELYSIA_CHUNKED_{collection_name.lower()}__")


class Chunker:
    def __init__(
        self,
        chunking_strategy: str = "fixed",
        num_tokens: int = 256,
        num_sentences: int = 5,
    ) -> None:
        self.chunking_strategy = chunking_strategy
        assert chunking_strategy in ["fixed", "sentences"]

        self.nlp = spacy.load("en_core_web_sm")

        self.num_tokens = num_tokens
        self.num_sentences = num_sentences

    def count_tokens(self, document: str) -> int:
        return len(self.nlp(document))

    def chunk_by_sentences(
        self,
        document: str,
        num_sentences: int | None = None,
        overlap_sentences: int = 1,
    ) -> tuple[list[str], list[tuple[int, int]]]:
        """
        Given a document (string), return the sentences as chunks and span annotations (start and end indices of chunks).
        Using spaCy to do sentence chunking.
        """
        if num_sentences is None:
            num_sentences = self.num_sentences

        if overlap_sentences >= num_sentences:
            print(
                f"Warning: overlap_sentences ({overlap_sentences}) is greater than num_sentences ({num_sentences}). Setting overlap to {num_sentences - 1}"
            )
            overlap_sentences = num_sentences - 1

        doc = self.nlp(document)
        sentences = list(doc.sents)  # Get sentence boundaries from spaCy

        span_annotations = []
        chunks = []

        i = 0
        while i < len(sentences):
            # Get chunk of num_sentences sentences
            chunk_sentences = sentences[i : i + num_sentences]
            if not chunk_sentences:
                break

            # Get start and end char positions
            start_char = chunk_sentences[0].start_char
            end_char = chunk_sentences[-1].end_char

            # Add chunk and its span annotation
            chunks.append(document[start_char:end_char])
            span_annotations.append((start_char, end_char))

            # Move forward but account for overlap
            i += num_sentences - overlap_sentences

        return chunks, span_annotations

    def chunk_by_tokens(
        self, document: str, num_tokens: int | None = None, overlap_tokens: int = 32
    ) -> tuple[list[str], list[tuple[int, int]]]:
        """
        Given a document (string), return the tokens as chunks and span annotations (start and end indices of chunks).
        Includes overlapping tokens between chunks for better context preservation.
        Uses spaCy for tokenization.
        """
        if num_tokens is None:
            num_tokens = self.num_tokens

        doc = self.nlp(document)
        tokens = list(doc)  # Get tokens from spaCy doc

        span_annotations = []
        chunks = []
        i = 0

        while i < len(tokens):
            # Find end index for current chunk
            end_idx = min(i + num_tokens, len(tokens))
            chunk_tokens = tokens[i:end_idx]

            # Get character spans for the chunk
            start_char = chunk_tokens[0].idx
            end_char = chunk_tokens[-1].idx + len(chunk_tokens[-1])

            # Add chunk and its span annotation
            chunks.append(document[start_char:end_char])
            span_annotations.append((start_char, end_char))

            # Move forward but account for overlap
            i += max(1, num_tokens - overlap_tokens)

        return chunks, span_annotations

    def chunk(self, document: str) -> tuple[list[str], list[tuple[int, int]]]:
        if self.chunking_strategy == "sentences":
            return self.chunk_by_sentences(document)
        elif self.chunking_strategy == "tokens":
            return self.chunk_by_tokens(document)
        else:
            raise ValueError(f"Invalid chunking strategy: {self.chunking_strategy}")


class AsyncCollectionChunker:
    def __init__(self, collection_name: str):
        self.collection_name = collection_name
        self.chunker = Chunker("sentences", num_sentences=5)

    async def create_chunked_reference(
        self, content_field: str, client_manager: ClientManager
    ) -> None:
        async with client_manager.connect_to_async_client() as client:
            chunked_collection = await self.get_chunked_collection(
                content_field, client
            )
            full_collection = client.collections.get(self.collection_name)

            # add reference to chunked collection
            try:
                await chunked_collection.config.add_reference(
                    ReferenceProperty(
                        name="fullDocument", target_collection=full_collection.name
                    )
                )
            except WeaviateInvalidInputError:
                # already exists
                pass

            # add reference to full collection
            try:
                await full_collection.config.add_reference(
                    ReferenceProperty(
                        name="isChunked", target_collection=chunked_collection.name
                    )
                )
            except WeaviateInvalidInputError:
                # already exists
                pass

    def get_chunked_collection_name(self) -> str:
        return f"ELYSIA_CHUNKED_{self.collection_name.lower()}__"

    async def chunked_collection_exists(self, client: WeaviateAsyncClient) -> bool:
        return await client.collections.exists(self.get_chunked_collection_name())

    async def get_vectoriser(
        self, content_field: str, client: WeaviateAsyncClient
    ) -> _VectorConfigCreate:
        collection_config = await client.collections.get(
            self.collection_name
        ).config.get()

        # see if any named vectors exclusively vectorise the content field
        if collection_config.vector_config is not None:
            for (
                named_vector_name,
                named_vector_config,
            ) in collection_config.vector_config.items():

                named_vectorizer = named_vector_config.vectorizer

                if (
                    named_vectorizer.source_properties is not None
                    and named_vectorizer.source_properties == [content_field]
                ):
                    try:
                        vectorizer = getattr(
                            Configure.Vectors,
                            named_vectorizer.vectorizer.replace("-", "_"),
                        )

                        valid_args = inspect.signature(vectorizer).parameters
                        return vectorizer(
                            name=named_vector_name,
                            source_properties=[content_field],
                            **{
                                arg: named_vectorizer.model[arg]
                                for arg in named_vectorizer.model
                                if arg in valid_args and arg != "vector_index_config"
                            },
                            vector_index_config=Configure.VectorIndex.hnsw(
                                quantizer=Configure.VectorIndex.Quantizer.sq()  # scalar quantization
                            ),
                        )
                    except AttributeError as e:
                        pass

        # if we haven't returned yet, try the overall vectoriser
        if collection_config.vectorizer_config is not None:
            try:
                vectorizer_name = (
                    collection_config.vectorizer_config.vectorizer.replace("-", "_")
                )
                vectorizer = getattr(Configure.Vectors, vectorizer_name)
                valid_args = inspect.signature(vectorizer).parameters

                return vectorizer(
                    **{
                        arg: collection_config.vectorizer_config.model[arg]
                        for arg in collection_config.vectorizer_config.model
                        if arg in valid_args and arg != "vector_index_config"
                    },
                    vector_index_config=Configure.VectorIndex.hnsw(
                        quantizer=Configure.VectorIndex.Quantizer.sq()  # scalar quantization
                    ),
                )
            except AttributeError as e:
                pass

        # otherwise use default weaviate embedding service
        return Configure.Vectors.text2vec_weaviate(
            dimensions=256,
            vector_index_config=Configure.VectorIndex.hnsw(
                quantizer=Configure.VectorIndex.Quantizer.sq()  # scalar quantization
            ),
        )

    async def get_chunked_collection(
        self, content_field: str, client: WeaviateAsyncClient
    ) -> CollectionAsync:
        if await client.collections.exists(self.get_chunked_collection_name()):
            return client.collections.get(self.get_chunked_collection_name())
        else:
            return await client.collections.create(
                self.get_chunked_collection_name(),
                properties=[
                    Property(name=content_field, data_type=DataType.TEXT),
                    Property(
                        name="chunk_spans",
                        data_type=DataType.INT_ARRAY,
                    ),
                ],
                references=[
                    ReferenceProperty(
                        name="fullDocument", target_collection=self.collection_name
                    )
                ],
                vector_config=await self.get_vectoriser(content_field, client),
            )

    def generate_uuids(
        self,
        chunks: list[str],
        spans: list[tuple[int, int]],
        content_field: str,
    ) -> list[str]:
        chunked_uuids = []
        for i, (chunk, span) in enumerate(zip(chunks, spans)):
            data_object = {content_field: chunk, "chunk_spans": span}
            chunked_uuids.append(generate_uuid5(data_object))
        return chunked_uuids

    async def chunk_single_object(
        self, object: Object, content_field: str
    ) -> tuple[list[str], list[tuple[int, int]], list[str]]:
        content_field_value: str = object.properties[content_field]
        chunks, spans = self.chunker.chunk(content_field_value)
        chunk_uuids = self.generate_uuids(chunks, spans, content_field)
        return chunks, spans, chunk_uuids

    async def chunk_objects_parallel(
        self,
        unchunked_objects: list[Object],
        unchunked_uuids: list[str],
        content_field: str,
    ) -> tuple[dict, dict, dict]:
        tasks = [
            self.chunk_single_object(obj, content_field) for obj in unchunked_objects
        ]

        results = await asyncio.gather(*tasks)

        original_uuid_to_chunks = {}
        original_uuid_to_spans = {}
        original_uuid_to_chunk_uuids = {}

        for uuid, (chunks, spans, chunk_uuids) in zip(unchunked_uuids, results):
            original_uuid_to_chunks[uuid] = chunks
            original_uuid_to_spans[uuid] = spans
            original_uuid_to_chunk_uuids[uuid] = chunk_uuids

        return (
            original_uuid_to_chunks,
            original_uuid_to_spans,
            original_uuid_to_chunk_uuids,
        )

    async def insert_chunks(
        self,
        chunked_collection: CollectionAsync,
        original_uuid_to_chunks: dict,
        original_uuid_to_spans: dict,
        original_uuid_to_chunk_uuids: dict,
        content_field: str,
    ) -> None:
        data_objects = []
        for original_uuid in original_uuid_to_chunks:
            for i, (chunk, span, uuid) in enumerate(
                zip(
                    original_uuid_to_chunks[original_uuid],
                    original_uuid_to_spans[original_uuid],
                    original_uuid_to_chunk_uuids[original_uuid],
                )
            ):
                data_objects.append(
                    DataObject(
                        properties={content_field: chunk, "chunk_spans": span},
                        uuid=uuid,
                        references={"fullDocument": original_uuid},
                    )
                )

        batch_size = 15
        data_objects_batches = [
            data_objects[i : i + batch_size]
            for i in range(0, len(data_objects), batch_size)
        ]

        tasks = [
            chunked_collection.data.insert_many(data_objects)
            for data_objects in data_objects_batches
        ]

        await asyncio.gather(*tasks)

    async def insert_references(
        self,
        full_collection: CollectionAsync,
        original_uuid_to_chunk_uuids: dict,
    ) -> None:
        references = []
        for original_uuid in original_uuid_to_chunk_uuids:
            for chunk_uuid in original_uuid_to_chunk_uuids[original_uuid]:
                references.append(
                    DataReference(
                        from_uuid=original_uuid,
                        from_property="isChunked",
                        to_uuid=chunk_uuid,
                    )
                )

        batch_size = 15
        references_batches = [
            references[i : i + batch_size]
            for i in range(0, len(references), batch_size)
        ]

        tasks = [
            full_collection.data.reference_add_many(references)
            for references in references_batches
        ]

        await asyncio.gather(*tasks)

    def get_chunked_objects(self, objects: list[Object]) -> list[str]:
        """
        Given a list of weaviate objects, find if a reference exists to a chunked object in the separate collection.
        Return the UUIDs of `objects` that have a reference to a chunked object (already have been chunked).
        """
        uuids = []
        for object in objects:
            if (
                object.references is not None
                and "isChunked" in object.references
                and len(object.references["isChunked"].objects) > 0
                and any(
                    [
                        o.collection == self.get_chunked_collection_name()
                        for o in object.references["isChunked"].objects
                    ]
                )
            ):
                uuids.append(str(object.uuid))
        return uuids

    async def __call__(
        self,
        weaviate_response: QueryReturn,
        content_field: str,
        client_manager: ClientManager,
    ) -> None:

        # get all UUIDs of current objects
        all_uuids = [str(object.uuid) for object in weaviate_response.objects]

        # get all UUIDs of chunked objects
        chunked_uuids = self.get_chunked_objects(weaviate_response.objects)

        # get all UUIDs of unchunked objects - these are the objects that need to be chunked
        unchunked_uuids = [
            str(uuid) for uuid in all_uuids if str(uuid) not in chunked_uuids
        ]

        # get all unchunked objects for chunking and inserting
        unchunked_objects = [
            object
            for object in weaviate_response.objects
            if str(object.uuid) in unchunked_uuids
        ]

        # if there are unchunked objects, chunk them
        if len(unchunked_objects) > 0:
            async with client_manager.connect_to_async_client() as client:
                # Get collections
                chunked_collection = await self.get_chunked_collection(
                    content_field, client
                )
                full_collection = client.collections.get(self.collection_name)

                (
                    original_uuid_to_chunks,
                    original_uuid_to_spans,
                    original_uuid_to_chunk_uuids,
                ) = await self.chunk_objects_parallel(
                    unchunked_objects, unchunked_uuids, content_field
                )

                # insert into weaviate
                await self.insert_chunks(
                    chunked_collection,
                    original_uuid_to_chunks,
                    original_uuid_to_spans,
                    original_uuid_to_chunk_uuids,
                    content_field,
                )
                await self.insert_references(
                    full_collection, original_uuid_to_chunk_uuids
                )



from elysia import Error, Tool, Result
import numpy as np
import matplotlib.pyplot as plt
from elysia.tools.visualisation.objects import (
    ScatterOrLineChart,
    ScatterOrLineDataPoints,
    ScatterOrLineDataPoint,
    ScatterOrLineYAxisData,
    ScatterOrLineChart,
    ChartResult,
)


class BasicLinearRegression(Tool):

    def __init__(self, logger, **kwargs):
        super().__init__(
            name="basic_linear_regression_tool",
            description="""
            Use this tool to perform linear regression between two numeric variables in the environment.
            This also returns a visualisation of the data.
            """,
            status="Running linear regression...",
            inputs={
                "environment_key": {
                    "description": (
                        "A single key (string) of the `environment` dictionary that will be used in the analysis. "
                        "Choose the most relevant key for the analysis according to the user prompt. "
                        "All objects under that key will be used to create the dataframe. "
                    ),
                    "required": True,
                    "type": str,
                    "default": None,
                },
                "x_variable_field": {
                    "description": (
                        "The independent variable for the regression. "
                        "Choose one field title within the `objects` underneath the specific `environment_key`. "
                    ),
                    "required": True,
                    "type": str,
                    "default": None,
                },
                "y_variable_field": {
                    "description": (
                        "The dependent variable for the regression. "
                        "Choose one single field within the `objects` underneath the specific `environment_key`. "
                    ),
                    "required": True,
                    "type": str,
                    "default": None,
                },
            },
            end=False,
        )

    async def __call__(
        self,
        tree_data,
        inputs,
        base_lm,
        complex_lm,
        client_manager,
        **kwargs,
    ):
        environment = tree_data.environment.environment
        environment_key = inputs["environment_key"]
        x_variable_field = inputs["x_variable_field"]
        y_variable_field = inputs["y_variable_field"]

        try:

            # initialise empty matrices to store all objects
            X = np.empty((0, 2))
            y = np.empty((0, 1))

            # iterate over all items in the environment
            for inner_key in environment[environment_key]:

                # convert all objects under this key to a matrix
                inner_X = np.array(
                    [
                        [obj[x_variable_field]]
                        for environment_list in environment[environment_key][inner_key]
                        for obj in environment_list["objects"]
                    ]
                )
                inner_X = np.hstack([np.ones((inner_X.shape[0], 1)), inner_X])
                X = np.vstack([X, inner_X])

                # convert all objects under this key to a matrix
                inner_y = np.array(
                    [
                        [obj[y_variable_field]]
                        for environment_list in environment[environment_key][inner_key]
                        for obj in environment_list["objects"]
                    ]
                )
                y = np.vstack([y, inner_y])

            # calculate the beta hat values via least squares
            beta_hat = np.linalg.inv(X.T @ X + 1e-10 * np.eye(X.shape[1])) @ X.T @ y
            beta_hat_dict = {
                "intercept": beta_hat[0],
                "slope": beta_hat[1],
            }
            pred_y = X @ beta_hat

            # plot the data
            fig, ax = plt.subplots()
            ax.scatter(X[:, 1], y)
            ax.plot(X[:, 1], pred_y, color="red")
            ax.set_title(
                f"Linear regression between {x_variable_field} and {y_variable_field}"
            )
            ax.set_xlabel(x_variable_field)
            ax.set_ylabel(y_variable_field)
            fig.show()

            # yield the result to the decision tree
            yield Result(
                objects=[
                    beta_hat_dict,
                ],
                metadata={
                    "x_variable_field": x_variable_field,
                    "y_variable_field": y_variable_field,
                },
                llm_message=(
                    "Completed linear regression analysis where: "
                    "    x variable: {x_variable_field} "
                    "    y variable: {y_variable_field} "
                    f"    intercept: {beta_hat_dict['intercept']} "
                    f"    slope: {beta_hat_dict['slope']} "
                    "And returned a visualisation of the data."
                ),
            )

            # create the chart payload
            x_axis = [ScatterOrLineDataPoint(value=x) for x in X[:, 1]]
            y_axis = [
                ScatterOrLineYAxisData(
                    label=y_variable_field,
                    kind="scatter",
                    data_points=[ScatterOrLineDataPoint(value=y) for y in y],
                ),
                ScatterOrLineYAxisData(
                    label="Line of best fit",
                    kind="line",
                    data_points=[ScatterOrLineDataPoint(value=y) for y in pred_y],
                ),
            ]
            chart = ScatterOrLineChart(
                title=f"Linear regression between {x_variable_field} and {y_variable_field}",
                description=f"Linear regression between {x_variable_field} and {y_variable_field}",
                x_axis_label=x_variable_field,
                y_axis_label=y_variable_field,
                data=ScatterOrLineDataPoints(x_axis=x_axis, y_axis=y_axis),
            )
            yield ChartResult([chart], "scatter_or_line")

        except Exception as e:
            yield Error(str(e))

    async def is_tool_available(self, tree_data, base_lm, complex_lm, client_manager):
        """
        Available when the 'query' task has been completed and it has added data to the environment.
        """
        return (
            "query" in tree_data.environment.environment
            and len(tree_data.environment.environment["query"]) > 0
        )



import elysia.api.custom_tools as custom_tools
from elysia import Tool
from typing import Dict, Type
from fastapi import APIRouter, Depends
from fastapi.responses import JSONResponse

from elysia.api.dependencies.common import get_user_manager
from elysia.api.services.user import UserManager
from elysia.api.core.log import logger
from elysia.tree.tree import Tree
from elysia.api.api_types import (
    AddToolToTreeData,
    RemoveToolFromTreeData,
    AddBranchToTreeData,
    RemoveBranchFromTreeData,
)
from elysia.api.services.tree import TreeManager

router = APIRouter()


def find_tool_classes() -> Dict[str, Type[Tool]]:
    """
    Find all Tool subclasses in the custom_tools module.

    Returns:
        (dict): Dictionary mapping class names to Tool subclass types
    """
    tool_classes = {}

    # Get all objects from the custom_tools module
    module_objects = dict(
        [
            (name, cls)
            for name, cls in custom_tools.__dict__.items()
            if isinstance(cls, type)
        ]
    )

    # Filter for Tool subclasses (excluding the base Tool class)
    for name, cls in module_objects.items():
        if issubclass(cls, Tool) and cls.__name__ != "Tool":
            tool_classes[name] = cls

    return tool_classes


def find_tool_metadata() -> Dict[str, Dict[str, str]]:
    """
    Find all tool metadata in the custom_tools module.

    Returns:
        (dict): Dictionary mapping tool names to tool metadata
    """
    tools = find_tool_classes()
    metadata = {}
    for name, cls in tools.items():
        metadata[name] = cls.get_metadata()
    return metadata


@router.get("/available")
async def get_available_tools():
    headers = {"Cache-Control": "no-cache"}

    try:
        return JSONResponse(
            content={"tools": find_tool_metadata(), "error": ""},
            status_code=200,
            headers=headers,
        )
    except Exception as e:
        logger.error(f"Error getting available tools: {str(e)}")
        return JSONResponse(
            content={"tools": {}, "error": str(e)},
            status_code=500,
            headers=headers,
        )


@router.post("/{user_id}/add")
async def add_tool_to_tree(
    user_id: str,
    data: AddToolToTreeData,
    user_manager: UserManager = Depends(get_user_manager),
):
    try:
        # get tool class
        tool_class = find_tool_classes()[data.tool_name]
        user = await user_manager.get_user_local(user_id)
        tree_manager: TreeManager = user["tree_manager"]

        # get tree and add tool
        for conversation_id in tree_manager.trees:
            tree: Tree = tree_manager.get_tree(conversation_id)
            tree.add_tool(tool_class, data.branch_id, from_tool_ids=data.from_tool_ids)

        # all trees should look the same, so return the most recent one
        return JSONResponse(content={"tree": tree.tree, "error": ""}, status_code=200)

    except Exception as e:
        logger.error(f"Error adding tool to tree: {str(e)}")
        return JSONResponse(content={"tree": {}, "error": str(e)}, status_code=500)


@router.post("/{user_id}/remove")
async def remove_tool_from_tree(
    user_id: str,
    data: RemoveToolFromTreeData,
    user_manager: UserManager = Depends(get_user_manager),
):
    try:
        # get tool class
        tool_class = find_tool_classes()[data.tool_name]

        # get tree and remove tool
        user = await user_manager.get_user_local(user_id)
        tree_manager: TreeManager = user["tree_manager"]
        for conversation_id in tree_manager.trees:
            tree: Tree = tree_manager.get_tree(conversation_id)
            tree.remove_tool(
                tool_class.get_metadata()["name"],  # type: ignore
                data.branch_id,
                from_tool_ids=data.from_tool_ids,
            )

        return JSONResponse(content={"tree": tree.tree, "error": ""}, status_code=200)
    except Exception as e:
        logger.error(f"Error removing tool from tree: {str(e)}")
        return JSONResponse(content={"tree": {}, "error": str(e)}, status_code=500)


@router.post("/{user_id}/add_branch")
async def add_branch_to_tree(
    user_id: str,
    data: AddBranchToTreeData,
    user_manager: UserManager = Depends(get_user_manager),
):
    try:
        # get tree and add branch
        user = await user_manager.get_user_local(user_id)
        tree_manager: TreeManager = user["tree_manager"]
        for conversation_id in tree_manager.trees:
            tree: Tree = tree_manager.get_tree(conversation_id)
            tree.add_branch(
                branch_id=data.id,
                instruction=data.instruction,
                description=data.description,
                root=data.root,
                from_branch_id=data.from_branch_id,
                from_tool_ids=data.from_tool_ids,
                status=data.status,
            )

        return JSONResponse(content={"tree": tree.tree, "error": ""}, status_code=200)
    except Exception as e:
        logger.error(f"Error adding branch to tree: {str(e)}")
        return JSONResponse(content={"tree": {}, "error": str(e)}, status_code=500)


@router.post("/{user_id}/remove_branch")
async def remove_branch_from_tree(
    user_id: str,
    data: RemoveBranchFromTreeData,
    user_manager: UserManager = Depends(get_user_manager),
):
    try:
        # get tree and remove branch
        user = await user_manager.get_user_local(user_id)
        tree_manager: TreeManager = user["tree_manager"]
        for conversation_id in tree_manager.trees:
            tree: Tree = tree_manager.get_tree(conversation_id)
            tree.remove_branch(data.id)

        return JSONResponse(content={"tree": tree.tree, "error": ""}, status_code=200)
    except Exception as e:
        logger.error(f"Error removing branch from tree: {str(e)}")
        return JSONResponse(content={"tree": {}, "error": str(e)}, status_code=500)



import os
from fastapi.responses import JSONResponse
from pathlib import Path
from logging import Logger
import json
import random
import pytest
from weaviate.util import generate_uuid5
from elysia.api.api_types import (
    InitialiseTreeData,
    SaveConfigUserData,
    SaveConfigTreeData,
    UpdateFrontendConfigData,
)
from elysia.api.dependencies.common import get_user_manager
from elysia.api.routes.init import initialise_tree, initialise_user
from elysia.api.routes.user_config import (
    load_config_user,
    delete_config,
    save_config_user,
    new_user_config,
    list_configs,
    update_frontend_config,
    get_models,
)
from elysia.api.routes.tree_config import (
    load_config_tree,
    change_config_tree,
    new_tree_config,
)
from elysia.tree.tree import Tree
from elysia.api.core.log import logger, set_log_level
from elysia.api.services.tree import TreeManager
from elysia.api.services.user import UserManager
from elysia.api.utils.config import FrontendConfig

set_log_level("CRITICAL")


def read_response(response: JSONResponse):
    return json.loads(response.body)


async def delete_config_after_completion(
    user_manager: UserManager, user_id: str, config_id: str
):
    client_manager = user_manager.users[user_id][
        "frontend_config"
    ].save_location_client_manager

    if client_manager.is_client:
        async with client_manager.connect_to_async_client() as client:
            if not await client.collections.exists("ELYSIA_CONFIG__"):
                return

            collection = client.collections.get("ELYSIA_CONFIG__")
            uuid = generate_uuid5(config_id)
            if await collection.data.exists(uuid):
                await collection.data.delete_by_id(uuid)

    if os.path.exists(f"elysia/api/user_configs/frontend_config_{user_id}.json"):
        os.remove(f"elysia/api/user_configs/frontend_config_{user_id}.json")


async def initialise_user_and_tree(user_id: str, conversation_id: str):
    user_manager = get_user_manager()

    response = await initialise_user(
        user_id,
        user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    response = await initialise_tree(
        user_id,
        conversation_id,
        InitialiseTreeData(
            low_memory=False,
        ),
        user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""


@pytest.mark.asyncio
async def test_set_save_location():
    """
    Setting the WCD_URL/WCD_API_KEY in the frontend config should be saved to the user config.
    """
    user_manager = get_user_manager()
    user_id = "test_user_set_save_location"
    conversation_id = "test_conversation_set_save_location"
    config_id = "test_config_set_save_location"
    config_name = "Test set save location"

    await initialise_user_and_tree(user_id, conversation_id)

    # the user should have a save location from the .env
    user = await user_manager.get_user_local(user_id)
    assert user["frontend_config"].save_location_wcd_url == os.getenv("WCD_URL")
    assert user["frontend_config"].save_location_wcd_api_key == os.getenv("WCD_API_KEY")

    # update the save location to an invalid one
    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name=config_name,
            default=True,
            config={},
            frontend_config={
                "save_location_wcd_url": "test_incorrect_wcd_url",
                "save_location_wcd_api_key": "test_incorrect_wcd_api_key",
            },
        ),
        user_manager=user_manager,
    )
    response = read_response(response)

    # but the user should have a new save location
    user = await user_manager.get_user_local(user_id)
    assert user["frontend_config"].save_location_wcd_url == "test_incorrect_wcd_url"
    assert (
        user["frontend_config"].save_location_wcd_api_key
        == "test_incorrect_wcd_api_key"
    )
    warning_found = False
    for warning in response["warnings"]:
        warning_found = "Could not connect to Weaviate" in warning
    assert warning_found

    # same if no api keys are provided
    # update the save location to an invalid one
    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name=config_name,
            default=True,
            config={},
            frontend_config={
                "save_location_wcd_url": "",
                "save_location_wcd_api_key": "",
            },
        ),
        user_manager=user_manager,
    )
    response = read_response(response)

    warning_found = False
    for warning in response["warnings"]:
        warning_found = "No valid destination for config save location found" in warning
    assert warning_found


@pytest.mark.asyncio
async def test_smart_setup_models_user():
    """
    Test that models are set based on supplied api keys.
    """
    user_manager = get_user_manager()
    user_id = "test_user_smart_setup_models_user"
    conversation_id = "test_conversation_smart_setup_models_user"

    await initialise_user_and_tree(user_id, conversation_id)

    # check that the tree manager has the correct settings
    tree_manager = user_manager.users[user_id]["tree_manager"]

    assert tree_manager.settings.BASE_MODEL is not None
    assert tree_manager.settings.COMPLEX_MODEL is not None
    assert tree_manager.settings.BASE_PROVIDER is not None
    assert tree_manager.settings.COMPLEX_PROVIDER is not None

    # should default use environment variables
    assert tree_manager.settings.WCD_URL == os.getenv("WCD_URL")
    assert tree_manager.settings.WCD_API_KEY == os.getenv("WCD_API_KEY")
    assert tree_manager.config.style != ""
    assert tree_manager.config.agent_description != ""
    assert tree_manager.config.end_goal != ""
    assert tree_manager.config.branch_initialisation != ""


@pytest.mark.asyncio
async def test_smart_setup_tree():
    """
    Test that models are set based on api keys in the .env for a tree.
    """
    user_manager = get_user_manager()
    user_id = "test_user_smart_setup_tree"
    conversation_id = "test_conversation_smart_setup_tree"

    await initialise_user_and_tree(user_id, conversation_id)
    tree: Tree = await user_manager.get_tree(user_id, conversation_id)

    # and the tree should have the correct models
    assert tree.settings.BASE_MODEL is not None
    assert tree.settings.COMPLEX_MODEL is not None
    assert tree.settings.BASE_PROVIDER is not None
    assert tree.settings.COMPLEX_PROVIDER is not None


@pytest.mark.asyncio
async def test_save_config():
    """
    Test saving a user config gets added to the Weaviate database.
    """
    user_manager = get_user_manager()
    user_id = "test_user_save_config"
    conversation_id = "test_conversation_save_config"
    config_id = f"test_config_conversation_save_config"
    config_name = "Test config save"

    await initialise_user_and_tree(user_id, conversation_id)

    # create a config and save
    user = await user_manager.get_user_local(user_id)
    tree_manager: TreeManager = user["tree_manager"]

    new_settings = {
        "WCD_URL": os.getenv("TESTING_WCD_URL"),
        "WCD_API_KEY": os.getenv("TESTING_WCD_API_KEY"),
    }

    # check that the first WCD_URL etc is the env one and not the new_settings ones
    config = user["tree_manager"].config.to_json()
    assert config["settings"]["WCD_URL"] == os.getenv("WCD_URL")
    assert config["settings"]["WCD_API_KEY"] == os.getenv("WCD_API_KEY")

    assert config["settings"]["WCD_URL"] != new_settings["WCD_URL"]
    assert config["settings"]["WCD_API_KEY"] != new_settings["WCD_API_KEY"]

    # change config
    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name=config_name,
            default=False,
            config={
                "settings": new_settings,
                "style": "Test style",
                "agent_description": "Test description",
                "end_goal": "Test end goal",
                "branch_initialisation": tree_manager.config.branch_initialisation,
            },
            frontend_config={},
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == "", response["error"]

    # verify the config was saved by listing all configs
    response = await list_configs(
        user_id=user_id,
        user_manager=user_manager,
    )

    response = read_response(response)
    assert response["error"] == "", response["error"]
    assert config_id in [c["config_id"] for c in response["configs"]]


@pytest.mark.asyncio
async def test_load_config_user():
    """
    Test loading a user config from the Weaviate database.
    """
    user_manager = get_user_manager()
    user_id = "test_user_load_config_user"
    conversation_id = "test_conversation_load_config_user"
    config_id = f"test_config_user"
    config_name = "Test config user"

    await initialise_user_and_tree(user_id, conversation_id)

    # create a config to save with custom settings
    settings = {
        "BASE_MODEL": "gpt-4o-mini",
        "COMPLEX_MODEL": "gpt-4o",
        "BASE_PROVIDER": "openai",
        "COMPLEX_PROVIDER": "openai",
    }
    custom_style = "Technical and precise."
    custom_agent_description = "You are a technical assistant focusing on details."
    custom_end_goal = "Provide technically accurate information to users."
    custom_branch_init = "multi_branch"

    # change user settings to these values
    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name=config_name,
            default=True,
            config={
                "settings": settings,
                "style": custom_style,
                "agent_description": custom_agent_description,
                "end_goal": custom_end_goal,
                "branch_initialisation": custom_branch_init,
            },
            frontend_config={},
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # now change settings to something else WITHOUT saving
    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name=config_name,
            default=False,
            config={
                "settings": {
                    "BASE_MODEL": "claude-3-haiku",
                    "COMPLEX_MODEL": "claude-3-opus",
                    "BASE_PROVIDER": "anthropic",
                    "COMPLEX_PROVIDER": "anthropic",
                },
                "style": "Conversational and friendly.",
                "agent_description": "You are a friendly assistant.",
                "end_goal": "Provide helpful and friendly responses.",
                "branch_initialisation": "one_branch",
            },
            frontend_config={
                "save_configs_to_weaviate": False,
            },
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # Load the config back for the user
    response = await load_config_user(
        user_id=user_id,
        config_id=config_id,
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # Check the user settings were updated back to the saved config
    user = await user_manager.get_user_local(user_id)
    tree_manager = user["tree_manager"]

    assert tree_manager.settings.BASE_MODEL == "gpt-4o-mini"
    assert tree_manager.settings.COMPLEX_MODEL == "gpt-4o"
    assert tree_manager.settings.BASE_PROVIDER == "openai"
    assert tree_manager.settings.COMPLEX_PROVIDER == "openai"
    assert tree_manager.config.style == custom_style
    assert tree_manager.config.agent_description == custom_agent_description
    assert tree_manager.config.end_goal == custom_end_goal


@pytest.mark.asyncio
async def test_load_config_tree():
    """
    Test loading a config for a tree from the Weaviate database.
    This should only update the tree settings, not the atlas or branch init.
    """
    user_manager = get_user_manager()
    user_id = "test_user_load_config_tree"
    config_id = "test_config_tree"
    conversation_id = "test_conversation_load_config_tree"
    config_name = "Test config tree"

    await initialise_user_and_tree(user_id, conversation_id)

    # Create a config to save
    settings = {
        "BASE_MODEL": "gpt-4-turbo",
        "COMPLEX_MODEL": "gpt-4-vision",
        "BASE_PROVIDER": "openai",
        "COMPLEX_PROVIDER": "openai",
    }
    custom_style = "Analytical and detailed."
    custom_agent_description = (
        "You are an analytical assistant providing detailed analyses."
    )
    custom_end_goal = "Provide thorough and analytical responses to queries."
    custom_branch_init = "empty"

    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name=config_name,
            default=True,
            config={
                "settings": settings,
                "style": custom_style,
                "agent_description": custom_agent_description,
                "end_goal": custom_end_goal,
                "branch_initialisation": custom_branch_init,
            },
            frontend_config={},
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # Load the config for the tree (just settings, not atlas or branch init)
    response = await load_config_tree(
        user_id=user_id,
        conversation_id=conversation_id,
        config_id=config_id,
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # Verify the tree settings were updated but not the atlas or branch init
    tree = await user_manager.get_tree(user_id, conversation_id)
    assert tree.settings.BASE_MODEL == "gpt-4-turbo"
    assert tree.settings.COMPLEX_MODEL == "gpt-4-vision"
    assert tree.settings.BASE_PROVIDER == "openai"
    assert tree.settings.COMPLEX_PROVIDER == "openai"

    # Load everything now
    response = await load_config_tree(
        user_id=user_id,
        conversation_id=conversation_id,
        config_id=config_id,
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # Verify everything changed
    tree = await user_manager.get_tree(user_id, conversation_id)
    assert tree.settings.BASE_MODEL == "gpt-4-turbo"
    assert tree.settings.COMPLEX_MODEL == "gpt-4-vision"
    assert tree.tree_data.atlas.style == custom_style
    assert tree.tree_data.atlas.agent_description == custom_agent_description
    assert tree.tree_data.atlas.end_goal == custom_end_goal
    # assert tree.branch_initialisation == custom_branch_init

    # User level settings should remain unchanged
    user = await user_manager.get_user_local(user_id)
    tree_manager = user["tree_manager"]
    assert tree_manager.settings.BASE_MODEL == "gpt-4-turbo"
    assert tree_manager.settings.COMPLEX_MODEL == "gpt-4-vision"
    assert tree_manager.config.style == custom_style


@pytest.mark.asyncio
async def test_list_configs():
    """
    Test listing all configs for a user from the Weaviate database.
    """

    user_manager = get_user_manager()
    user_id = "test_user_list_configs"
    conversation_id = "test_conversation_list_configs"
    config_ids = [
        f"test_config_list_{i}_{random.randint(0, 1000000)}" for i in range(3)
    ]
    await initialise_user_and_tree(user_id, conversation_id)

    # Save multiple configs
    for i, config_id in enumerate(config_ids):
        config = SaveConfigUserData(
            name=f"Test config list {i}",
            default=False,
            config={
                "settings": {
                    "BASE_MODEL": f"model-{i}",
                    "COMPLEX_MODEL": f"complex-model-{i}",
                    "BASE_PROVIDER": "test-provider",
                    "COMPLEX_PROVIDER": "test-provider",
                },
                "style": f"Style {i}",
                "agent_description": f"Agent description {i}",
                "end_goal": f"End goal {i}",
                "branch_initialisation": "one_branch",
            },
            frontend_config={},
        )

        response = await save_config_user(
            user_id=user_id,
            config_id=config_id,
            data=config,
            user_manager=get_user_manager(),
        )
        response = read_response(response)
        assert response["error"] == ""

    response = await list_configs(
        user_id=user_id,
        user_manager=get_user_manager(),
    )
    response = read_response(response)
    assert response["error"] == ""

    for config_id in config_ids:
        assert config_id in [c["config_id"] for c in response["configs"]]


@pytest.mark.asyncio
async def test_config_save_load_workflow():
    """
    Test the config save/load workflow within a Weaviate database.
    """
    user_manager = get_user_manager()
    user_id_1 = "test_user_config_workflow_1"
    conversation_id_1 = "test_conversation_workflow_1"
    conversation_id_2 = "test_conversation_workflow_2"
    config_id = "test_shared_config_workflow"
    config_name = "Test shared config workflow"

    await initialise_user_and_tree(user_id_1, conversation_id_1)

    # Initialize the user with custom settings
    custom_settings = {
        "BASE_MODEL": "gpt-4o-mini",
        "COMPLEX_MODEL": "gpt-4o",
        "BASE_PROVIDER": "openai",
        "COMPLEX_PROVIDER": "openai",
    }
    custom_style = "Original style"
    custom_agent_description = "Original agent description"
    custom_end_goal = "Original end goal"

    # Apply the config
    user1 = await user_manager.get_user_local(user_id_1)
    tree_manager1: TreeManager = user1["tree_manager"]

    response = await save_config_user(
        user_id=user_id_1,
        config_id=config_id,
        data=SaveConfigUserData(
            name=config_name,
            default=True,
            config={
                "settings": tree_manager1.settings.to_json(),
                "style": custom_style,
                "agent_description": custom_agent_description,
                "end_goal": custom_end_goal,
                "branch_initialisation": tree_manager1.config.branch_initialisation,
            },
            frontend_config={
                "save_configs_to_weaviate": False,
            },
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # update frontend config so that configs are saved to weaviate
    response = await save_config_user(
        user_id=user_id_1,
        config_id=config_id,
        data=SaveConfigUserData(
            name=config_name,
            default=True,
            config={
                "settings": custom_settings,
                "style": custom_style,
                "agent_description": custom_agent_description,
                "end_goal": custom_end_goal,
                "branch_initialisation": tree_manager1.config.branch_initialisation,
            },
            frontend_config={
                "save_location_wcd_url": os.getenv("WCD_URL"),
                "save_location_wcd_api_key": os.getenv("WCD_API_KEY"),
                "save_trees_to_weaviate": False,
                "save_configs_to_weaviate": True,
            },
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # list configs
    response = await list_configs(
        user_id=user_id_1,
        user_manager=user_manager,
    )
    list_response1 = read_response(response)
    assert config_id in [c["config_id"] for c in list_response1["configs"]]

    # change the config only in-memory
    response = await save_config_user(
        user_id=user_id_1,
        config_id=config_id,
        data=SaveConfigUserData(
            name=config_name,
            default=True,
            config={
                "settings": {
                    "BASE_MODEL": "claude-3-haiku",
                    "BASE_PROVIDER": "anthropic",
                },
                "style": "New incorrect style",
                "agent_description": "New incorrect agent description",
                "end_goal": "New incorrect end goal",
                "branch_initialisation": tree_manager1.config.branch_initialisation,
            },
            frontend_config={
                "save_configs_to_weaviate": False,
            },
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # check the new config options are applied
    user1 = await user_manager.get_user_local(user_id_1)
    tree_manager1: TreeManager = user1["tree_manager"]
    assert tree_manager1.settings.BASE_MODEL == "claude-3-haiku"
    assert tree_manager1.settings.BASE_PROVIDER == "anthropic"
    assert tree_manager1.config.style == "New incorrect style"
    assert tree_manager1.config.agent_description == "New incorrect agent description"
    assert tree_manager1.config.end_goal == "New incorrect end goal"

    # Now load the config to the user
    response = await load_config_user(
        user_id=user_id_1,
        config_id=config_id,
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # Now verify the new settings were loaded
    user1 = await user_manager.get_user_local(user_id_1)
    tree_manager1: TreeManager = user1["tree_manager"]
    assert tree_manager1.settings.BASE_MODEL == "gpt-4o-mini"
    assert tree_manager1.config.style == "Original style"
    assert tree_manager1.config.agent_description == "Original agent description"
    assert tree_manager1.config.end_goal == "Original end goal"

    user1_frontend_config = user1["frontend_config"]
    assert user1_frontend_config.save_location_wcd_url == os.getenv("WCD_URL")
    assert user1_frontend_config.save_location_wcd_api_key == os.getenv("WCD_API_KEY")
    assert user1_frontend_config.config["save_trees_to_weaviate"] == False

    # Make a new tree for this user
    response = await initialise_tree(
        user_id=user_id_1,
        conversation_id=conversation_id_2,
        data=InitialiseTreeData(
            low_memory=True,
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # dont save
    response = await update_frontend_config(
        user_id=user_id_1,
        data=UpdateFrontendConfigData(
            config={"save_configs_to_weaviate": False},
        ),
        user_manager=user_manager,
    )

    # change the config globally for the user
    response = await save_config_user(
        user_id=user_id_1,
        config_id=config_id,
        data=SaveConfigUserData(
            name=config_name,
            default=True,
            config={
                "settings": {
                    "BASE_MODEL": "claude-3-5-sonnet",
                    "BASE_PROVIDER": "anthropic",
                },
                "style": "New updated style",
                "agent_description": "New updated agent description",
                "end_goal": "New updated end goal",
                "branch_initialisation": tree_manager1.config.branch_initialisation,
            },
            frontend_config={},
        ),
        user_manager=user_manager,
    )

    # Changes to user should propagate to the tree (not used its own settings)
    tree2: Tree = await user_manager.get_tree(user_id_1, conversation_id_2)
    assert tree2.settings.BASE_MODEL == "claude-3-5-sonnet"  # this one changes
    assert tree2.tree_data.atlas.style == "New updated style"
    assert tree2.tree_data.atlas.agent_description == "New updated agent description"
    assert tree2.tree_data.atlas.end_goal == "New updated end goal"

    # Load the config to the new tree 1 directly
    response = await load_config_tree(
        user_id=user_id_1,
        conversation_id=conversation_id_2,
        config_id=config_id,
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # Verify tree has the ORIGINAL settings (from the save/load)
    tree2 = await user_manager.get_tree(user_id_1, conversation_id_2)
    assert tree2.settings.BASE_MODEL == "gpt-4o-mini"
    assert tree2.settings.COMPLEX_MODEL == "gpt-4o"
    assert tree2.tree_data.atlas.style == "Original style"
    assert tree2.tree_data.atlas.agent_description == "Original agent description"
    assert tree2.tree_data.atlas.end_goal == "Original end goal"


@pytest.mark.asyncio
async def test_delete_config():
    """
    Test deleting a config from the Weaviate database.
    """
    user_manager = get_user_manager()
    user_id = "test_user_delete_config"
    conversation_id = "test_conversation_delete_config"
    config_id = "test_delete_config"
    config_name = "Test delete config"
    await initialise_user_and_tree(user_id, conversation_id)

    # create a config and save
    user = await user_manager.get_user_local(user_id)
    tree_manager: TreeManager = user["tree_manager"]

    # change config
    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name=config_name,
            default=False,
            config={
                "settings": tree_manager.settings.to_json(),
                "style": tree_manager.config.style,
                "agent_description": tree_manager.config.agent_description,
                "end_goal": tree_manager.config.end_goal,
                "branch_initialisation": tree_manager.config.branch_initialisation,
            },
            frontend_config={},
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # verify the config was saved by listing all configs
    response = await list_configs(
        user_id=user_id,
        user_manager=user_manager,
    )

    response = read_response(response)
    assert response["error"] == ""
    assert config_id in [c["config_id"] for c in response["configs"]]

    # delete the config
    response = await delete_config(
        user_id=user_id,
        config_id=config_id,
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # verify the config was deleted by listing all configs
    response = await list_configs(
        user_id=user_id,
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""
    assert config_id not in [c["config_id"] for c in response["configs"]]



from fastapi import APIRouter, Depends
from fastapi.responses import JSONResponse

from elysia.api.api_types import (
    AddFeedbackData,
    RemoveFeedbackData,
)
from elysia.api.dependencies.common import get_user_manager
from elysia.api.services.user import UserManager
from elysia.api.utils.feedback import (
    create_feedback,
    feedback_metadata,
    remove_feedback,
)
from elysia.api.core.log import logger
from elysia.util.client import ClientManager

router = APIRouter()


@router.post("/add")
async def run_add_feedback(
    data: AddFeedbackData, user_manager: UserManager = Depends(get_user_manager)
):
    logger.debug(f"/add_feedback API request received")
    logger.debug(f"User ID: {data.user_id}")
    logger.debug(f"Conversation ID: {data.conversation_id}")
    logger.debug(f"Query ID: {data.query_id}")
    logger.debug(f"Feedback: {data.feedback}")

    tree = await user_manager.get_tree(data.user_id, data.conversation_id)
    user = user_manager.users[data.user_id]
    client_manager: ClientManager = user["client_manager"]

    async with client_manager.connect_to_async_client() as client:
        try:
            await create_feedback(
                data.user_id,
                data.conversation_id,
                data.query_id,
                data.feedback,
                tree,
                client,
            )
        except Exception as e:
            return JSONResponse(content={"error": str(e)}, status_code=500)

    return JSONResponse(content={"error": ""}, status_code=200)


@router.post("/remove")
async def run_remove_feedback(
    data: RemoveFeedbackData, user_manager: UserManager = Depends(get_user_manager)
):
    logger.debug(f"/remove_feedback API request received")
    logger.debug(f"User ID: {data.user_id}")
    logger.debug(f"Conversation ID: {data.conversation_id}")
    logger.debug(f"Query ID: {data.query_id}")

    user = user_manager.users[data.user_id]
    client_manager: ClientManager = user["client_manager"]

    async with client_manager.connect_to_async_client() as client:
        try:
            await remove_feedback(
                data.user_id, data.conversation_id, data.query_id, client
            )
            return JSONResponse(content={"error": ""}, status_code=200)
        except Exception as e:
            return JSONResponse(content={"error": str(e)}, status_code=500)


@router.get("/metadata/{user_id}")
async def run_feedback_metadata(
    user_id: str, user_manager: UserManager = Depends(get_user_manager)
):
    logger.debug(f"/feedback_metadata API request received")
    logger.debug(f"User ID: {user_id}")

    user = user_manager.users[user_id]
    client_manager: ClientManager = user["client_manager"]

    async with client_manager.connect_to_async_client() as client:
        metadata = await feedback_metadata(client, user_id)

    return JSONResponse(
        content={
            "error": "",
            **metadata,
        },
        status_code=200,
    )
